# -*- coding: UTF-8 -*-

from sklearn.tree import DecisionTreeClassifier, DecisionTreeRegressor
from sklearn.model_selection import RandomizedSearchCV
from sklearn.metrics import make_scorer
import numpy as np

from skater.model.base import ModelType
from skater.core.visualizer.tree_visualizer import plot_tree, tree_to_text

from skater.util.logger import build_logger
from skater.util.logger import _INFO, _DEBUG
from skater.util import exceptions
from skater.data import DataManager


class TreeSurrogate(object):
    """ :: Experimental :: The implementation is currently experimental and might change in future.
    The idea of using TreeSurrogates as means for explaining a model's(Oracle or the base model)
    learned decision policies(for inductive learning tasks) is inspired by the work of Mark W. Craven
    described as the TREPAN algorithm. In this explanation learning hypothesis, the base estimator(Oracle)
    could be any form of supervised learning predictive models. The explanations are approximated using
    DecisionTrees(both for Classification/Regression) by learning decision boundaries similar to that learned by
    the Oracle(predictions from the base model are used for learning the DecisionTree representation).
    The implementation also generates a fidelity score to quantify tree based surrogate model's
    approximation to the Oracle. Ideally, the score should be 0 for truthful explanation
    both globally and locally.

    Parameters
    ----------
    oracle : InMemory instance type
        model instance having access to the base estimator(InMemory/DeployedModel).
        Currently, only InMemory is supported.
    splitter : str (default="best")
        Strategy used to split at each the node. Supported strategies("best" or "random").
    max_depth : int (default=None)
        Defines the maximum depth of a tree. If 'None' then nodes are expanded till all leaves are \
        pure or contain less than min_samples_split samples.
        Deeper trees are prone to be more expensive and tend to over-fit.
        Pruning is a technique which could be applied to avoid over-fitting.
    min_samples_split : int/float (default=2)
        Defines the minimum number of samples required to split an internal node:

        - int, specifies the minimum number of samples
        - float, then represents a percentage. Minimum number of samples is computed as \
          `ceil(min_samples_split*n_samples)`

    min_samples_leaf : int/float (default=1)
        Defines requirement for a leaf node. The minimum number of samples needed to be a leaf node:

        - int, specifies the minimum number of samples
        - float, then represents a percentage. Minimum number of samples is computed as \
          `ceil(min_samples_split*n_samples)

    min_weight_fraction_leaf : float (default=0.0)
        Defines requirement for a leaf node. The minimum weight percentage of the sum total of the weights of \
        all input samples.
    max_features : int, float, string or None (default=None)
        Defines number of features to consider for the best possible split:

        - None, all specified features are used (oracle.feature_names)
        - int, uses specified values as `max_features` at each split.
        - float, as a percentage. Value for split is computed as `int(max_features * n_features)`.
        - "auto", `max_features=sqrt(n_features)`.
        - "sqrt", `max_features=sqrt(n_features)`.
        - "log2", `max_features=log2(n_features)`.

    seed : int, (default=None)
        seed for random number generator
    max_leaf_nodes : int or None (default=None)
        TreeSurrogates are constructed top-down in best first manner(best decrease in relative impurity).
        If None, results in maximum possible number of leaf nodes. This tends to over-fitting.
    min_impurity_decrease : float (default=0.0)
        Tree node is considered for splitting if relative decrease in impurity is >= `min_impurity_decrease`.
    class_weight : dict, list of dicts, str ("balanced" or None) (default="balanced")
        Weights associated with classes for handling data imbalance:

        - None, all classes have equal weights
        - "balanced", adjusts the class weights automatically. Weights are assigned inversely proportional \
          to class frequencies ``n_samples / (n_classes * np.bincount(y))``

    presort : bool (default=False)
        Sorts the data before building surrogates trees to find the best splits. When dealing with larger datasets, \
        setting it to True might result in increasing computation time because of the pre sorting operation.
    impurity_threshold : float (default=0.01)
        Specifies the acceptable disparity between the Oracle and TreeSurrogates. The higher the difference between \
        the Oracle and TreeSurrogate less faithful are the explanations generated.

    Attributes
    ----------
    oracle : skater.model.local_model.InMemoryModel
        The fitted base model with the prediction function
    feature_names: list of str
        Names of the features considered.
    estimator_ : DecisionTreeClassifier/DecisionTreeRegressor
        The Surrogate estimator.
    estimator_type_ : str
        Surrogate estimator type ("classifier" or "regressor").
    best_score_ : numpy.float64
        Surrogate estimator's best score post pre-pruning.
    scorer_name_ : str
        Scorer used for optimizing the surrogate estimator

    Examples
    --------
    >>> from skater.core.explanations import Interpretation
    >>> from skater.model import InMemoryModel
    >>> from skater.util.logger import _INFO
    >>> interpreter = Interpretation(X_train, feature_names=iris.feature_names)
    >>> model_inst = InMemoryModel(clf.predict, examples=X_train, model_type='classifier', unique_values=[0, 1],
    >>>                       feature_names=iris.feature_names, target_names=iris.target_names, log_level=_INFO)
    >>> # Using the interpreter instance invoke call to the TreeSurrogate
    >>> surrogate_explainer = interpreter.tree_surrogate(oracle=model_inst, seed=5)
    >>> surrogate_explainer.fit(X_train, y_train, use_oracle=True, prune='post', scorer_type='default')
    >>> surrogate_explainer.plot_global_decisions(colors=['coral', 'lightsteelblue','darkkhaki'],
    >>>                                          file_name='simple_tree_pre.png')
    >>> show_in_notebook('simple_tree_pre.png', width=400, height=300)

    References
    ----------
    .. [1] Mark W. Craven(1996) EXTRACTING COMPREHENSIBLE MODELS FROM TRAINED NEURAL NETWORKS
           (http://ftp.cs.wisc.edu/machine-learning/shavlik-group/craven.thesis.pdf)
    .. [2] Mark W. Craven and Jude W. Shavlik(NIPS, 96). Extracting Thee-Structured Representations of Thained Networks
           (https://papers.nips.cc/paper/1152-extracting-tree-structured-representations-of-trained-networks.pdf)
    .. [3] DecisionTreeClassifier: http://scikit-learn.org/stable/modules/generated/sklearn.tree.DecisionTreeClassifier.html
    .. [4] DecisionTreeRegressor: http://scikit-learn.org/stable/modules/generated/sklearn.tree.DecisionTreeRegressor.html
    """
    __name__ = "TreeSurrogate"

    def __init__(self, oracle=None, splitter='best', max_depth=None, min_samples_split=2,
                 min_samples_leaf=1, min_weight_fraction_leaf=0.0, max_features=None, seed=None, max_leaf_nodes=None,
                 min_impurity_decrease=0.0, min_impurity_split=None, class_weight="balanced",
                 presort=False, impurity_threshold=0.01):

        if not isinstance(oracle, ModelType):
            raise exceptions.ModelError("Incorrect estimator used, create one with skater.model.local.InMemoryModel")
        self.oracle = oracle
        self.logger = build_logger(oracle.logger.level, __name__)
        self.__model_type = None
        self.feature_names = oracle.feature_names
        self.class_names = oracle.target_names
        self.impurity_threshold = impurity_threshold
        self.criterion_types = {'classifier': {'criterion': ['gini', 'entropy']},
                                'regressor': {'criterion': ['mse', 'friedman_mse', 'mae']}}
        self.splitter_types = ['best', 'random']
        self.splitter = splitter if any(splitter in item for item in self.splitter_types) else 'best'
        self.seed = seed
        self.__model_type = oracle.model_type
        self.__scorer_name = None
        self.__best_score = None

        # TODO validate the parameters based on estimator type
        if self.__model_type == 'classifier':
            est = DecisionTreeClassifier(splitter=self.splitter, max_depth=max_depth,
                                         min_samples_split=min_samples_split, min_samples_leaf=min_samples_leaf,
                                         min_weight_fraction_leaf=min_weight_fraction_leaf,
                                         max_features=max_features, random_state=seed,
                                         max_leaf_nodes=max_leaf_nodes,
                                         min_impurity_decrease=min_impurity_decrease,
                                         class_weight=class_weight, presort=presort)
        elif self.__model_type == 'regressor':
            est = DecisionTreeRegressor(splitter=self.splitter, max_depth=None,
                                        min_samples_split=min_samples_split,
                                        min_samples_leaf=min_samples_leaf,
                                        min_weight_fraction_leaf=min_weight_fraction_leaf,
                                        max_features=max_features,
                                        random_state=seed, max_leaf_nodes=max_leaf_nodes,
                                        min_impurity_split=min_impurity_split, presort=presort)
        else:
            raise exceptions.ModelError("Model type not supported. Supported options types{'classifier', 'regressor'}")
        self.__model = est
        self.__pred_func = lambda X, prob: self.__model.predict(X) if prob is False else self.__model.predict_proba(X)


    @staticmethod
    def __optimizer_condition(o_s, new_s, scoring_type, threshold):
        # if optimizing on a loss function then the type is decreasing vs optimizing on a model metric which is increasing
        if scoring_type == 'decreasing':
            return round(o_s, 3) + threshold >= round(new_s, 3)
        else:
            return round(o_s, 3) - threshold <= round(new_s, 3)


    def _post_pruning(self, X, Y, scorer_type, impurity_threshold, needs_prob=False):
        self.__model.fit(X, Y)
        y_pred = self.__pred_func(X, needs_prob)
        # makes sense for classification use-case, be cautious when enabling for regression
        self.logger.debug("Unique Labels in ground truth provided {}".format(np.unique(Y)))
        if needs_prob is False:
            self.logger.debug("Unique Labels in predictions generated {}".format(np.unique(y_pred)))
        else:
            self.logger.debug("Probability scoring is enabled min:{}/max:{}".format(np.min(y_pred), np.max(y_pred)))

        scorer = self.oracle.scorers.get_scorer_function(scorer_type=scorer_type)
        self.logger.info("Scorer used {}".format(scorer.name))
        original_score = scorer(Y, y_pred)
        self.logger.info("original score using base model {}".format(original_score))

        tree = self.__model.tree_
        no_of_nodes = tree.node_count
        tree_leaf = -1  # value to identify a leaf node in a tree

        removed_node_index = []
        for index in range(no_of_nodes):
            current_left, current_right = tree.children_left[index], tree.children_right[index]
            if tree.children_left[index] != tree_leaf or tree.children_right[index] != tree_leaf:
                tree.children_left[index], tree.children_right[index] = -1, -1
                new_score = scorer(Y, self.__pred_func(X, needs_prob))
                self.logger.debug("new score generate {}".format(new_score))

                if TreeSurrogate.__optimizer_condition(original_score, new_score, scorer.type, impurity_threshold):
                    removed_node_index.append(index)
                    self.logger.debug("Removed nodes: (index:{}-->[left node: {}, right node: {}])"
                                      .format(index, current_left, current_right))
                else:
                    tree.children_left[index], tree.children_right[index] = current_left, current_right
                    self.logger.debug("Added index {} back".format(index))
        self.logger.info("Summary: childrens of the following nodes are removed {}".format(removed_node_index))


    def _pre_pruning(self, X, Y, scorer_type, cv=5, n_iter_search=10, n_jobs=1, param_grid=None, verbose=False):
        default_grid = {
            "criterion": self.criterion_types[self.__model_type]['criterion'],
            "max_depth": [2, 4, 6, 8, 10, 12],  # helps in reducing the depth of the tree
            "min_samples_leaf": [2, 4],  # restrict the minimum number of samples in a leaf
            "max_leaf_nodes": [2, 4, 6, 8, 10]  # reduce the number of leaf nodes
        }
        search_space = param_grid if param_grid is not None else default_grid
        self.logger.debug("Default search space used for CV : {}".format(search_space))
        # Cost function aiming to optimize(Total Cost) = measure of fit + measure of complexity
        # References for pruning:
        # 1. http://scikit-learn.org/stable/modules/model_evaluation.html#scoring-parameter
        # 2. https://www.coursera.org/lecture/ml-classification/optional-pruning-decision-trees-to-avoid-overfitting-qvf6v
        # Using Randomize Search here to prune the trees to improve readability without
        # comprising on model's performance
        scorer = self.oracle.scorers.get_scorer_function(scorer_type=scorer_type)
        self.logger.info("Scorer used {}".format(scorer.name))
        scorering_func = make_scorer(scorer, greater_is_better=scorer.type)
        verbose_level = 0 if verbose is False else 4
        random_search_estimator = RandomizedSearchCV(estimator=self.__model, cv=cv, param_distributions=search_space,
                                                     scoring=scorering_func, n_iter=n_iter_search, n_jobs=n_jobs,
                                                     random_state=self.seed, verbose=verbose_level)
        # train a surrogate DT
        random_search_estimator.fit(X, Y)
        # access the best estimator
        self.__model = random_search_estimator.best_estimator_
        self.__best_score = random_search_estimator.best_score_


    def fit(self, X, Y, use_oracle=True, prune='post', cv=5, n_iter_search=10,
            scorer_type='default', n_jobs=1, param_grid=None, impurity_threshold=0.01, verbose=False):
        """ Learn an approximate representation by constructing a Decision Tree based on the results retrieved by
        querying the Oracle(base model). Instances used for training should belong to the base learners instance space.

        Parameters
        ----------
        X : numpy.ndarray, pandas.DataFrame
            Training input samples
        Y : numpy.ndarray, target values(ground truth)
        use_oracle : bool (defaul=True)
            Use of Oracle, helps the Surrogate model train on the decision boundaries learned by the base model. \
            The closer the surrogate model is to the Oracle, more faithful are the explanations.

              - True, builds a surrogate model against the predictions of the base model(Oracle).
              - False, learns an interpretable tree based model using the supplied training examples and ground truth.

        prune : None, str (default="post")
            Pruning is a useful technique to control the complexity of the tree (keeping the trees comprehensive \
            and interpretable) without compromising on model's accuracy. Avoiding to build large and deep trees \
            also helps in preventing over-fitting.

              - "pre"
              Also known as forward/online pruning. This pruning process uses a termination \
              condition(high and low thresholds) to prematurely terminate some of the branches and nodes.
              Cross Validation is applied to measure the goodness of the fit while the tree is pruned.

              - "pos"
              Also known as backward pruning. The pruning process is applied post the construction of the \
              tree using the specified model parameters. This involves reducing the branches and nodes using \
              a cost function. The current implementation support cost optimization using \
              Model's scoring metrics(e.g. r2, log-loss, f1, ...).

        cv : int, (default=5)
            Randomized cross validation used only for 'pre-pruning' right now.
        n_iter_search : int (default=10)
            Number of parameter setting combinations that are sampled for pre-pruning.
        scorer_type : str (default="default")
        n_jobs : int (default=1)
            Number of jobs to run in parallel.
        param_grid : dict
            Dictionary of parameters to specify the termination condition for pre-pruning.
        impurity_threshold : float (default=0.01)
            Specifies acceptable performance drop when using Tree based surrogates to replicate the decision policies
            learned by the Oracle
        verbose : bool (default=False)
            Helps control the verbosity.

        References
        ----------
        .. [1] Nikita Patel and Saurabh Upadhyay(2012)
               Study of Various Decision Tree Pruning Methods with their Empirical Comparison in WEKA
               (https://pdfs.semanticscholar.org/025b/8c109c38dc115024e97eb0ede5ea873fffdb.pdf)
        """

        if verbose:
            self.logger.setLevel(_DEBUG)
        else:
            self.logger.setLevel(_INFO)
        # DataManager does type checking as well
        dm = DataManager(X, Y)
        X, Y = dm.X, dm.y
        # Below is an anti-pattern but had to use it. Should fix it in the long term
        y_hat_original = self.oracle._execute(X)

        # TODO: Revisit the check on using probability or class labels
        if use_oracle and self.oracle.probability:
            y_train = np.array(list(map(np.argmax, y_hat_original)))
        elif use_oracle:
            y_train = y_hat_original
        else:
            # this is when y_train is being passed and the desire is to build an interpretable tree based model
            y_train = Y

        if prune is None:
            self.logger.info("No pruning applied ...")
            self.__model.fit(X, y_train)
        elif prune == 'pre':
            # apply randomized cross validation for pruning
            self.logger.info("pre pruning applied ...")
            self._pre_pruning(X, y_train, scorer_type, cv, n_iter_search, n_jobs, param_grid, verbose)
        else:
            self.logger.info("post pruning applied ...")
            # Since, this is post pruning, we first learn a model
            # and then try to prune the tree controling the model's score using the impurity_threshold
            self._post_pruning(X, y_train, scorer_type, impurity_threshold, needs_prob=self.oracle.probability)
        y_hat_surrogate = self.__pred_func(X, self.oracle.probability)
        self.logger.info('Done generating prediction using the surrogate, shape {}'.format(y_hat_surrogate.shape))

        # Default metrics:
        # {Classification: if probability score used --> cross entropy(log-loss) else --> F1 score}
        # {Regression: Mean Absolute Error (MAE)}
        scorer = self.oracle.scorers.get_scorer_function(scorer_type=scorer_type)
        self.__scorer_name = scorer.name

        oracle_score = round(scorer(Y, y_hat_original), 3)
        # Since surrogate model is build against the base model's(Oracle's) predicted
        # behavior y_true=y_train
        surrogate_score = round(scorer(y_train, y_hat_surrogate), 3)
        self.logger.info('Done scoring, surrogate score {}; oracle score {}'.format(surrogate_score, oracle_score))

        impurity_score = round(oracle_score - surrogate_score, 3)
        if impurity_score > self.impurity_threshold:
            self.logger.warning('impurity score: {} of the surrogate model is higher than the impurity threshold: {}. '
                                'The higher the impurity score, lower is the fidelity/faithfulness '
                                'of the surrogate model'.format(impurity_score, impurity_threshold))
        return impurity_score


    @property
    def estimator_(self):
        """ Learned approximate surrogate estimator
        """
        return self.__model


    @property
    def estimator_type_(self):
        """ Estimator type
        """
        return self.__model_type


    @property
    def best_score_(self):
        """ Best score post pre-pruning
        """
        return self.__best_score


    @property
    def scorer_name_(self):
        """ Cost function used for optimization
        """
        return self.__scorer_name


    def predict(self, X, prob_score=False):
        """ Predict for input X
        """
        predict_values = self.__model.predict(X)
        predict_prob_values = self.__model.predict_proba(X) if prob_score is True else None
        return predict_values if predict_prob_values is None else predict_prob_values


    def plot_global_decisions(self, colors=None, enable_node_id=True, random_state=0, file_name="interpretable_tree.png",
                              show_img=False, fig_size=(20, 8)):
        """ Visualizes the decision policies of the surrogate tree.
        """
        graph_inst = plot_tree(self.__model, self.__model_type, feature_names=self.feature_names, color_list=colors,
                               class_names=self.class_names, enable_node_id=enable_node_id, seed=random_state)
        f_name = "interpretable_tree.png" if file_name is None else file_name
        graph_inst.write_png(f_name)

        try:
            import matplotlib
            matplotlib.use('agg')
            import matplotlib.pyplot as plt
        except ImportError:
            raise exceptions.MatplotlibUnavailableError("Matplotlib is required but unavailable on the system.")
        except RuntimeError:
            raise exceptions.MatplotlibDisplayError("Matplotlib unable to open display")

        if show_img:
            plt.rcParams["figure.figsize"] = fig_size
            img = plt.imread(f_name)
            if self.__model_type == 'regressor':
                cax = plt.imshow(img, cmap=plt.cm.get_cmap(graph_inst.get_colorscheme()))
                plt.colorbar(cax)
            else:
                plt.imshow(img)
        return graph_inst


    def decisions_as_txt(self, scope='global', X=None):
        """ Retrieve the decision policies as text
        """
        tree_to_text(self.__model, self.feature_names, self.__model_type, scope, X)



import unittest
import numpy as np
from skimage.util.shape import view_as_windows

from skater.util.image_ops import view_windows
from skater.util.image_ops import load_image
from skater.util.image_ops import standard_scaler
from skater.util.image_ops import normalize
from skater.util.image_ops import flip_orientation


class TestImageOps(unittest.TestCase):

    def setUp(self):
        pass


    def test_loading_imgs(self):
        img = load_image('skater/tests/data/pandas.jpg', 299, 299)
        self.assertEquals(img.shape, (299, 299, 3))
        img = load_image('skater/tests/data/pandas.jpg', 200, 299)
        self.assertEquals(img.shape, (200, 299, 3))


    def test_view_windows(self):
        input_matrix = np.arange(10 * 10).reshape(10, 10)
        padded_input = np.pad(input_matrix, ((0, 2), (0, 2)), mode='constant', constant_values=0)
        input_view = view_as_windows(padded_input, (4, 4), 4)
        # apply rolling window with padding to handle corner cases
        input_view_modified = view_windows(padded_input, (4, 4), 4)
        self.assertEquals(input_view.shape, input_view_modified.shape)


    def test_standard_scaler(self):
        x = np.asarray([1, 4, 5, 6, 6, 2, 3])
        x_new = standard_scaler(x)
        self.assertEquals(round(x_new.std(), 1), 1)


    def test_normalization(self):
        x = np.asarray([1, 4, 5, 6, 6, 2, 3])
        x_new = normalize(x)
        self.assertEquals((np.min(x_new), np.max(x_new)), (0, 1))


    def test_image_ops(self):
        # This is not an extensive test, but at-least a start
        X = np.arange(8).reshape((2, 2, 2))
        original_value = X[0, 0, 0]
        post_horizontal_flip = flip_orientation(X)
        self.assertEquals(original_value != post_horizontal_flip[0, 0, 0], True)


if __name__ == '__main__':
    suite = unittest.TestLoader().loadTestsFromTestCase(TestImageOps)
    unittest.TextTestRunner(verbosity=2).run(suite)



# -*- coding: UTF-8 -*-

from skimage.filters import roberts, sobel
import numpy as np
from skater.util.exceptions import MatplotlibUnavailableError
try:
    import matplotlib.pyplot as plt
except ImportError:
    raise (MatplotlibUnavailableError("Matplotlib is required but unavailable on your system."))

from skater.util.image_ops import normalize

# helper function to enable or disable matplotlib access
_enable_axis = lambda ax, flag: ax.axis("off") if flag is True else ax.axis("on")


def visualize(relevance_score, original_input_img=None, edge_detector_type='sobel', cmap='bwr', axis=plt,
              percentile=100, alpha_edges=0.8, alpha_bgcolor=1, disable_axis=True):

    dx, dy = 0.01, 0.01
    xx = np.arange(0.0, relevance_score.shape[1], dx)
    yy = np.arange(0.0, relevance_score.shape[0], dy)

    x_min, x_max, y_min, y_max = np.amin(xx), np.amax(xx), np.amin(yy), np.amax(yy)
    extent = x_min, x_max, y_min, y_max
    xi_cmap = plt.cm.gray
    xi_cmap.set_bad(alpha=0)

    edges = _edge_detection(original_input_img, edge_detector_type) if original_input_img is not None else None

    # draw the edges of the image before overlaying rest of the image
    if edges is not None:
        axis.imshow(edges, extent=extent, interpolation='nearest', cmap=xi_cmap, alpha=alpha_edges)

    abs_max = np.percentile(np.abs(relevance_score), percentile)
    abs_min = abs_max

    relevance_score = relevance_score[:, :, 0] if len(relevance_score.shape) == 3 else relevance_score
    # Plot the image with relevance scores
    axis.imshow(relevance_score, extent=extent, interpolation='nearest', cmap=cmap,
                vmin=-abs_min, vmax=abs_max, alpha=alpha_bgcolor)
    _enable_axis(axis, disable_axis)
    return axis


def _edge_detection(original_input_img=None, edge_detector_alg='sobel'):
    # Normalize the input image to (0,1)
    xi = normalize(original_input_img)
    xi_greyscale = xi if len(xi.shape) == 2 else np.mean(xi, axis=-1)
    # Applying edge detection ( Roberts or Sobel edge detection )
    # Reference: http://scikit-image.org/docs/0.11.x/auto_examples/plot_edge_filter.html
    edge_detector = {'robert': roberts, 'sobel': sobel}
    return edge_detector[edge_detector_alg](xi_greyscale)



from __future__ import division
import numpy as np
import pandas as pd
from sklearn.metrics.pairwise import cosine_distances
import ast
from functools import partial

from ..util.logger import build_logger
from ..util import exceptions
from ..util.static_types import StaticTypes
from ..util.dataops import add_column_numpy_array, allocate_samples_to_bins, flatten

__all__ = ['DataManager']


class DataManager(object):
    """Module for passing around data to interpretation objects
    """

    # Todo: we can probably remove some of the keys from data_info, and have properties
    # executed as pure functions for easy to access metadata, such as n_rows, etc

    _n_rows = 'n_rows'
    _dim = 'dim'
    _feature_info = 'feature_info'
    _dtypes = 'dtypes'

    __attribute_keys__ = [_n_rows, _dim, _feature_info, _dtypes]
    __datatypes__ = (pd.DataFrame, pd.Series, np.ndarray)

    def _check_X(self, X):
        if not isinstance(X, self.__datatypes__):
            err_msg = 'Invalid Data: expected data to be a numpy array or pandas dataframe but got ' \
                      '{}'.format(type(X))
            raise(exceptions.DataSetError(err_msg))
        ndim = len(X.shape)
        self.logger.debug("__init__ data.shape: {}".format(X.shape))

        if ndim == 1:
            X = X[:, np.newaxis]

        elif ndim >= 3:
            err_msg = "Invalid Data: expected data to be 1 or 2 dimensions, " \
                      "Data.shape: {}".format(ndim)
            raise(exceptions.DataSetError(err_msg))
        return X


    def _check_y(self, y, X):
        """
        convert y to ndarray

        If y is a dataframe:
            return df.values as ndarray
        if y is a series
            return series.values as ndarray
        if y is ndarray:
            return self
        if y is a list:
            return as ndarray
        :param y:
        :param X:
        :return:
        """

        if y is None:
            return None

        assert len(X) == len(y), \
            "len(X) = {0} does not equal len(y) = {1}".format(len(X), len(y))

        if isinstance(y, (pd.DataFrame, pd.Series)):
            return y.values
        elif isinstance(y, np.ndarray):
            return y
        elif isinstance(y, list):
            return np.array(y)
        else:
            raise ValueError("Unrecognized type for y: {}".format(type(y)))


    def __init__(self, X, y=None, feature_names=None, index=None, log_level=30):
        """
        The abstraction around using, accessing, sampling data for interpretation purposes.
        Used by interpretation objects to grab data, collect samples, and handle
        feature names and row indices.

        Parameters
        ----------
            X: 1D/2D numpy array, or pandas DataFrame
                raw data
            y: 1D/2D numpy array, or pandas DataFrame
                ground truth labels for X
            feature_names: iterable of feature names
                Optional keyword containing names of features.
            index: iterable of row names
                Optional keyword containing names of indexes (rows).

        """

        # create logger
        self._log_level = log_level
        self.logger = build_logger(log_level, __name__)

        self.X = self._check_X(X)
        self.y = self._check_y(y, self.X)
        self.data_type = type(self.X)
        self.metastore = None

        self.logger.debug("after transform X.shape: {}".format(self.X.shape))

        if isinstance(self.X, pd.DataFrame):
            if feature_names is None:
                feature_names = self.X.columns.values
            if index is None:
                index = range(self.n_rows)
            self.X.index = index

        elif isinstance(self.X, np.ndarray):
            if feature_names is None:
                feature_names = range(self.X.shape[1])
            if index is None:
                index = range(self.n_rows)

        else:
            raise(ValueError("Invalid: currently we only support {}"
                             "If you would like support for additional data structures let us "
                             "know!".format(self.__datatypes__)))

        self.feature_ids = list(feature_names)
        self.index = list(index)
        self.data_info = {attr: None for attr in self.__attribute_keys__}


    def generate_grid(self, feature_ids, grid_resolution=100, grid_range=(.05, .95)):
        """
        Generates a grid of values on which to compute pdp. For each feature xi, for value
        yj of xi, we will fix xi = yj for every observation in X.

        Parameters
        ----------
            feature_ids(list):
                Feature names for which we'll generate a grid. Must be contained
                by self.feature_ids

            grid_resolution(int):
                The number of unique values to choose for each feature.

            grid_range(tuple):
                The percentile bounds of the grid. For instance, (.05, .95) corresponds to
                the 5th and 95th percentiles, respectively.

        Returns
        -------
        grid(numpy.ndarray): 	There are as many rows as there are feature_ids
                                There are as many columns as specified by grid_resolution
        """

        if not all(i >= 0 and i <= 1 for i in grid_range):
            err_msg = "Grid range values must be between 0 and 1 but got:" \
                      "{}".format(grid_range)
            raise(exceptions.MalformedGridRangeError(err_msg))

        if not isinstance(grid_resolution, int) and grid_resolution > 0:
            err_msg = "Grid resolution {} is not a positive integer".format(grid_resolution)
            raise(exceptions.MalformedGridRangeError(err_msg))

        if not all(feature_id in self.feature_ids for feature_id in feature_ids):
            missing_features = []
            for feature_id in feature_ids:
                if feature_id not in self.feature_ids:
                    missing_features.append(feature_id)
            err_msg = "Feature ids {} not found in DataManager.feature_ids".format(missing_features)
            raise(KeyError(err_msg))

        grid_range = [x * 100 for x in grid_range]
        bins = np.linspace(*grid_range, num=grid_resolution).tolist()
        grid = []
        for feature_id in feature_ids:
            data = self[feature_id]
            info = self.feature_info[feature_id]
            # if a feature is categorical (non numeric) or
            # has a small number of unique values, we'll just
            # supply unique values for the grid
            if info['unique'] < grid_resolution or info['numeric'] is False:
                vals = np.unique(data)
            else:
                vals = np.unique(np.percentile(data, bins))
            grid.append(vals)
        grid = np.array(grid)
        grid_shape = [(1, i) for i in [row.shape[0] for row in grid]]
        self.logger.info('Generated grid of shape {}'.format(grid_shape))
        return grid


    def sync_metadata(self):
        self.data_info[self._n_rows] = self.n_rows
        self.data_info[self._dim] = self.dim
        self.data_info[self._dtypes] = self.dtypes
        self.data_info[self._feature_info] = self._calculate_feature_info()


    def _calculate_n_rows(self):
        return self.X.shape[0]


    def _calculate_dim(self):
        return self.X.shape[1]

    @property
    def values(self):
        if self.data_type == pd.DataFrame:
            result = self.X.values
        else:
            result = self.X
        return result


    @property
    def dtypes(self):
        return pd.DataFrame(self.X, columns=self.feature_ids, index=self.index).dtypes


    @property
    def shape(self):
        return self.X.shape


    @property
    def n_rows(self):
        return self.shape[0]


    @property
    def dim(self):
        return self.shape[1]


    def _calculate_feature_info(self):
        feature_info = {}
        for feature in self.feature_ids:
            x = self[feature]
            samples = self.generate_column_sample(feature, n_samples=10)
            samples_are_numeric = map(StaticTypes.data_types.is_numeric, np.array(samples))
            is_numeric = all(samples_are_numeric)
            feature_info[feature] = {
                'type': self.dtypes.loc[feature],
                'unique': len(np.unique(x)),
                'numeric': is_numeric
            }
        return feature_info

    @property
    def feature_info(self):
        if self.data_info[self._feature_info] is None:
            self.data_info[self._feature_info] = self._calculate_feature_info()
        return self.data_info[self._feature_info]


    def _build_metastore(self):

        medians = np.median(self.X, axis=0).reshape(1, self.dim)

        # how far each data point is from the global median
        dists = cosine_distances(self.X, Y=medians).reshape(-1)

        sorted_index = [self.index[i] for i in dists.argsort()]

        return {'sorted_index': sorted_index}

    def __repr__(self):
        return self.X.__repr__()

    def __iter__(self):
        for i in self.feature_ids:
            yield i


    def __setitem__(self, key, newval):
        if issubclass(self.data_type, pd.DataFrame) or issubclass(self.data_type, pd.Series):
            self.__setcolumn_pandas__(key, newval)
        elif issubclass(self.data_type, np.ndarray):
            self.__setcolumn_ndarray__(key, newval)
        else:
            raise ValueError("Can't set item for data of type {}".format(self.data_type))
        self.sync_metadata()


    def __setcolumn_pandas__(self, i, newval):
        """if you passed in a pandas dataframe, it has columns which are strings."""
        self.X[i] = newval


    def __setcolumn_ndarray__(self, i, newval):
        """if you passed in a pandas dataframe, it has columns which are strings."""

        if i in self.feature_ids:
            idx = self.feature_ids.index(i)
            self.X[:, idx] = newval
        else:
            self.X = add_column_numpy_array(self.X, newval)
            self.feature_ids.append(i)


    def __getitem__(self, key):
        if issubclass(self.data_type, pd.DataFrame) or issubclass(self.data_type, pd.Series):
            return self.__getitem_pandas__(key)
        elif issubclass(self.data_type, np.ndarray):
            return self.__getitem_ndarray__(key)
        else:
            raise ValueError("Can't get item for data of type {}".format(self.data_type))


    def __getitem_pandas__(self, i):
        """if you passed in a pandas dataframe, it has columns which are strings."""
        return self.X[i]


    def __getitem_ndarray__(self, i):
        """if you passed in a pandas dataframe, it has columns which are strings."""
        if StaticTypes.data_types.return_data_type(i) == StaticTypes.output_types.iterable:
            idx = [self.feature_ids.index(j) for j in i]
            return self.X[:, idx]
        elif StaticTypes.data_types.is_string(i) or StaticTypes.data_types.is_numeric(i):
            idx = self.feature_ids.index(i)
            return self.X[:, idx]
        else:
            raise(ValueError("Unrecongized index type: {}. This should not happen, "
                             "submit a issue here: "
                             "https://github.com/datascienceinc/Skater/issues"
                             .format(type(i))))


    def __getrows__(self, idx):
        if self.data_type == pd.DataFrame:
            return self.__getrows_pandas__(idx)
        elif self.data_type == np.ndarray:
            return self.__getrows_ndarray__(idx)
        else:
            raise ValueError("Can't get rows for data of type {}".format(self.data_type))


    def __getrows_pandas__(self, idx):
        """if you passed in a pandas dataframe, it has columns which are strings."""
        if StaticTypes.data_types.return_data_type(idx) == StaticTypes.output_types.iterable:
            i = [self.index.index(i) for i in idx]
        else:
            i = [self.index[idx]]
        return self.X.iloc[i]


    def __getrows_ndarray__(self, idx):
        """if you passed in a pandas dataframe, it has columns which are strings."""
        i = [self.index.index(i) for i in idx]
        return self.X[i]


    def generate_sample(self, sample=True, include_y=False, strategy='random-choice', n_samples=1000,
                        replace=True, bin_count=50):
        """ Method for generating data from the dataset.

        Parameters
        -----------
            sample : boolean
                If False, we'll take the full dataset, otherwise we'll sample.

            include_y: boolean (default=False)

            strategy: string (default='random-choice')
                Supported strategy types 'random-choice', 'uniform-from-percentile', 'uniform-over-similarity-ranks'

            n_samples : int (default=1000)
                Specifies the number of samples to return. Only implemented if strategy is "random-choice".

            replace : boolean (default=True)
                Bool for sampling with or without replacement

            bin_count : int
                If strategy is "uniform-over-similarity-ranks", then this is the number
                of samples to take from each discrete rank.
        """

        __strategy_types__ = ['random-choice', 'uniform-from-percentile', 'uniform-over-similarity-ranks']

        bin_count, samples_per_bin = allocate_samples_to_bins(n_samples, ideal_bin_count=bin_count)
        arg_dict = {
            'sample': sample,
            'strategy': strategy,
            'n_samples': n_samples,
            'replace': replace,
            'samples_per_bin': samples_per_bin,
            'bin_count': bin_count
        }
        self.logger.debug("Generating sample with args:\n {}".format(arg_dict))

        if not sample:
            idx = self.index

        if strategy == 'random-choice':
            idx = np.random.choice(self.index, size=n_samples, replace=replace)

        elif strategy == 'uniform-from-percentile':
            raise(NotImplementedError("We havent coded this yet."))

        elif strategy == 'uniform-over-similarity-ranks':
            sorted_index = self._build_metastore()['sorted_index']
            range_of_indices = list(range(len(sorted_index)))

            def aggregator(samples_per_bin, list_of_indicies):
                n = samples_per_bin[aggregator.count]
                result = str(np.random.choice(list_of_indicies, size=n).tolist())
                aggregator.count += 1
                return result

            aggregator.count = 0
            agg = partial(aggregator, samples_per_bin)

            cuts = pd.qcut(range_of_indices, [i / bin_count for i in range(bin_count + 1)])
            cuts = pd.Series(cuts).reset_index()
            indices = cuts.groupby(0)['index'].aggregate(agg).apply(lambda x: ast.literal_eval(x)).values
            indices = flatten(indices)
            idx = [self.index[i] for i in indices]
        else:
            raise ValueError("Strategy {0} not recognized, currently supported strategies: {1}".format(
                strategy,
                __strategy_types__
            ))
        if include_y:
            return self.__getrows__(idx), self._labels_by_index(idx)
        else:
            return self.__getrows__(idx)


    def generate_column_sample(self, feature_id, *args, **kwargs):
        """Sample a single feature from the data set.

        Parameters
        ----------
        feature_id: hashable
            name of the feature to sample. If no feature names were passed, then
            the features are accessible via their column index.

        """
        dm = DataManager(self[feature_id],
                         feature_names=[feature_id],
                         index=self.index)
        return dm.generate_sample(*args, **kwargs)


    def set_index(self, index):
        self.index = index
        if self.data_type in (pd.DataFrame, pd.Series):
            self.X.index = index


    def _labels_by_index(self, data_index):
        """ Method for grabbing labels associated with given indices.
        """
        # we coerce self.index to a list, so this is fine:
        numeric_index = [self.index.index(i) for i in data_index]

        # do we need to coerce labels to a particular data type?
        return self.y[numeric_index]


    @classmethod
    def _check_input(cls, dataset):
        """
        Ensures that dataset is pandas dataframe, and dataset is not empty
        :param dataset: skater.__datatypes__
        :return:
        """
        if not isinstance(dataset, (pd.DataFrame)):
            err_msg = "dataset must be a pandas.DataFrame"
            raise exceptions.DataSetError(err_msg)

        if len(dataset) == 0:
            err_msg = "dataset is empty"
            raise exceptions.DataSetError(err_msg)



import numpy as np
from skater.util.text_ops import cleaner


def _handling_ngrams_wts(original_feat_dict):
    # Currently, when feature dictionary contains continuous sequences such as 2-gram/3-gram etc. sequences as features,
    # a short term solution is to further split them into additional keys.
    # e.g. {'stay ball dropped': 0.5} will be split into a new dict as {'stay':0.5, 'ball': 0.5, 'dropped': 0.5}
    # TODO: this is just a temporary solution for handling n-grams, figure out a better solution
    for k in list(original_feat_dict.keys()):
        additional_keys = k.split()
    for a_k in additional_keys:
        if a_k in original_feat_dict:
            original_feat_dict[a_k] += original_feat_dict[a_k]
        else:
            original_feat_dict[a_k] = original_feat_dict[k]
    new_dict = original_feat_dict
    return new_dict


def relevance_wt_assigner(raw_txt, wts_as_dict):
    # normalize score by absolute max value
    if isinstance(wts_as_dict, dict):
        feature_scores = _handling_ngrams_wts(wts_as_dict)
        max_wt = np.max(np.abs(list(feature_scores.values())))
        wts_as_dict = {word: feature_scores[word] / max_wt for word in feature_scores.keys()}
        # transform dict into list of tuples (word, relevance_wts)
        # TODO: look into removing the below occurring side effect
        relevance_wts = []
        for word in raw_txt.split():
            # Clean up the raw word for irregularities
            word_cleaned_as_key = cleaner(word)
            if word_cleaned_as_key in wts_as_dict.keys():
                relevance_wts.append((word, wts_as_dict[word_cleaned_as_key]))
            else:
                relevance_wts.append((word, None))
    else:
        raise Exception('relevance wts currently needs to be as dict')
    return relevance_wts



# -*- coding: UTF-8 -*-
from skater.core.local_interpretation.dnni.initializer import Initializer
from skater.util.exceptions import TensorflowUnavailableError
try:
    import tensorflow as tf
except ImportError:
    raise (TensorflowUnavailableError("TensorFlow binaries are not installed"))
import numpy as np
from skater.util.logger import build_logger
from skater.util.logger import _INFO


class BaseGradientMethod(Initializer):
    """
    Base class for gradient-based relevance computation

    Reference
    - https://github.com/marcoancona/DeepExplain/blob/master/deepexplain/tensorflow/methods.py
    """

    __name__ = "BaseGradientMethod"
    logger = build_logger(_INFO, __name__)

    def _default_relevance_score(self):
        BaseGradientMethod.logger.debug("Computing default relevance score...")
        return tf.gradients(self.output_tensor, self.input_tensor)


    def _run(self):
        BaseGradientMethod.logger.info("Executing operations ...")
        relevance_scores = self._default_relevance_score()
        results = self._session_run(relevance_scores, self.samples)
        return results[0]


    @classmethod
    def _non_linear_grad(cls, op, grad):
        BaseGradientMethod.logger.debug("Computing gradient with activation type {}".format(op.type))
        return cls._original_grad(op, grad)


class LRP(BaseGradientMethod):
    """ LRP is technique to decompose the prediction(output) of a deep neural networks(DNNs) by computing relevance at
    each layer in a backward pass. Current implementation is computed using backpropagation by applying change rule on
    a modified gradient function. LRP could be implemented in different ways.
    This version implements the epsilon-LRP(Eq (58) as stated in [1] or Eq (2) in [2].
    Epsilon acts as a numerical stabilizer.

    References
    ----------
    .. [1] Bach S, Binder A, Montavon G, Klauschen F, Müller K-R, Samek W (2015)
       On Pixel-Wise Explanations for Non-Linear Classifier Decisions by Layer-Wise Relevance Propagation.
       PLoS ONE 10(7): e0130140. https://doi.org/10.1371/journal.pone.0130140
    .. [2] Ancona M, Ceolini E, Öztireli C, Gross M:
           Towards better understanding of gradient-based attribution methods for Deep Neural Networks. ICLR, 2018
    """
    __name__ = "LRP"
    _eps = None
    logger = build_logger(_INFO, __name__)

    def __init__(self, output_tensor, input_tensor, samples, session, epsilon=1e-4):
        super(LRP, self).__init__(output_tensor, input_tensor, samples, session)
        assert epsilon > 0.0, 'LRP epsilon must be > 0'
        LRP._eps = epsilon
        LRP.logger.info("Epsilon value: {}".format(LRP._eps))


    def _default_relevance_score(self):
        # computing dot product of the feature wts of the input data and the gradients of the prediction label
        return [g * x for g, x in
                zip(tf.gradients(self.output_tensor, self.input_tensor), [self.input_tensor])]


    @classmethod
    def _non_linear_grad(cls, op, grad):
        LRP.logger.debug("Computing non-linear gradient with activation type {}".format(op.type))
        op_out = op.outputs[0]
        op_in = op.inputs[0]
        stabilizer_epsilon = cls._eps * tf.sign(op_in)
        op_in += stabilizer_epsilon
        return grad * op_out / op_in


class IntegratedGradients(BaseGradientMethod):
    """ Integrated Gradient is a relevance scoring algorithm for Deep network based on final predictions to its input
    features. The algorithm statisfies two fundamental axioms related to relevance/attribution computation,
     1.Sensitivity : For every input and baseline, if the change in one feature causes the prediction to change,
     then the that feature should have non-zero relevance score

     2.Implementation Invariance : Compute relevance(attribution) should be identical for functionally equivalent
     networks.

    References
    ----------
    .. [1] Sundararajan, Mukund, Taly, Ankur, Yan, Qiqi (ICML, 2017).
    .. Axiomatic Attribution for Deep Networks (http://arxiv.org/abs/1703.01365)
    .. [2] Ancona M, Ceolini E, Öztireli C, Gross M (ICLR, 2018).
    .. Towards better understanding of gradient-based attribution methods for Deep Neural Networks.
    .. [3] Taly, Ankur(2017) http://theory.stanford.edu/~ataly/Talks/sri_attribution_talk_jun_2017.pdf
    """
    __name__ = "IntegratedGradients"
    logger = build_logger(_INFO, __name__)

    def __init__(self, output_tensor, input_tensor, samples, session, steps=100, baseline=None):
        super(IntegratedGradients, self).__init__(output_tensor, input_tensor, samples, session)
        self.steps = steps
        self.baseline = self._validate_baseline(baseline)
        # validate baseline, if baseline == None, _validate_baseline assigns the default baseline
        # black image or zero embedding vector for text is used as a default baseline, as suggested in the paper
        # Mukund Sundararajan, Ankir Taly, Qibi Yan. Axiomatic Attribution for Deep Networks(ICML2017)
        self._validate_baseline(self.baseline)


    def _run(self):
        IntegratedGradients.logger.info("Executing operations to compute relevance using Integrated Gradient")
        t_grad = self._default_relevance_score()
        gradient = None
        alpha_list = list(np.linspace(start=1. / self.steps, stop=1.0, num=self.steps))
        for alpha in alpha_list:
            xs_scaled = (self.samples - self.baseline) * alpha
            # compute the gradient for each alpha value
            _scores = self._session_run(t_grad, xs_scaled)
            gradient = _scores if gradient is None else [g + a for g, a in zip(gradient, _scores)]

        results = [(x - b) * (g / self.steps) for g, x, b in zip(gradient, [self.samples], [self.baseline])]
        return results[0]



# Reference: http://python-packaging.readthedocs.io/en/latest/dependencies.html

__title__ = 'skater'
__version__ = '1.1.2'
__summary__ = 'Model Interpretation Library'
__uri__ = 'https://github.com/datascienceinc/skater/'
__author__ = 'Aaron Kramer, Pramit Choudhary, DataScience.com Team, and contributors'
__maintainer__ = 'Aaron Kramer, Pramit Choudhary and community'
__email__ = 'dstm@datascience.com'
__license__ = 'MIT'
__docs__ = 'https://datascienceinc.github.io/Skater/'



import unittest

from sklearn.datasets import make_moons
from sklearn.tree import DecisionTreeClassifier
from sklearn.model_selection._split import train_test_split

from skater.model import InMemoryModel


class TestScorer(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        X, y = make_moons(1000, noise=0.5)
        cls.X_train, cls.X_test, cls.y_train, cls.y_test = train_test_split(X, y)
        cls.classifier_est = DecisionTreeClassifier(max_depth=5)
        cls.classifier_est.fit(cls.X_train, cls.y_train)


    def test_compute_default_scores(self):
        # For classification default scorer is weighted F1-score
        model_inst = InMemoryModel(self.classifier_est.predict, examples=self.X_train,
                                   model_type='classifier', unique_values=[0, 1, 2])
        scorer = model_inst.scorers.get_scorer_function(scorer_type='default')
        self.assertEqual(scorer.name == 'f1-score', True)

        scorer = model_inst.scorers.get_scorer_function(scorer_type='f1')
        self.assertEqual(scorer.name == 'f1-score', True)

        y_hat = self.classifier_est.predict(self.X_test)
        value = scorer(self.y_test, y_hat, average='weighted')
        self.assertEquals(value > 0, True)


    def test_compute_log_loss(self):
        model_inst = InMemoryModel(self.classifier_est.predict_proba, examples=self.X_train, probability=True,
                                   model_type='classifier')
        scorer = model_inst.scorers.get_scorer_function(scorer_type='default')
        self.assertEqual(scorer.name == 'cross-entropy', True)

        scorer = model_inst.scorers.get_scorer_function(scorer_type='cross_entropy')
        self.assertEqual(scorer.name == 'cross-entropy', True)

        y_hat = self.classifier_est.predict_proba(self.X_test)
        value = scorer(self.y_test, y_hat)
        self.assertEquals(value > 0, True)


if __name__ == '__main__':
    suite = unittest.TestLoader().loadTestsFromTestCase(TestScorer)
    unittest.TextTestRunner(verbosity=2).run(suite)



import numpy as np
from sklearn.model_selection import StratifiedKFold
from sklearn import metrics



def compute_validation_curve(estimator, n_folds, x=None, y=None, param_name=None, param_range=None):
    """
    This is a convenient function to compute train and test scores for varying parameter values. Its a simplified form
    of parameter optimization but with one variable. This might come handy once the optimized parameters are
    determined but one wants to investigate and understand the impact of different parameters of the estimator.
    Also, provides training scores for visualization allowing to investigate whether the model is overfitting or
    underfitting.

    Reference:
    ----------
    .. [1] http://scikit-learn.org/stable/modules/learning_curve.html#validation-curves-plotting-scores-to-evaluate-models
    ( sklearn's implementation is tied to the sklearn estimator type hence not allowing comparison with other models
    from other frameworks )
    """
    # TODO: make it scalable
    c_v = StratifiedKFold(n_splits=n_folds)
    param_result_train = []
    param_result_test = []
    for v in param_range:
        cv_result_train = []
        cv_result_test = []
        for train_idx, test_idx in c_v.split(x, y):
            param_dict = {param_name: v}
            estimator.set_params(param_dict)
            estimator.fit(x.iloc[train_idx], y.iloc[train_idx])

            train_scores = estimator.predict_prob(x.iloc[train_idx])
            train_metric = roc_auc_score(y.iloc[train_idx], train_scores[1], pos_label=1)

            test_scores = estimator.predict_proba(x.iloc[test_idx])
            test_metric = roc_auc_score(y.iloc[test_idx], test_scores[1], pos_label=1)
            cv_result_train.append(train_metric[0])
            cv_result_test.append(test_metric[0])
        param_result_train.append(cv_result_train)
        param_result_test.append(cv_result_test)
    return np.array(param_result_train), np.array(param_result_test)


def roc_auc_score(y_true, y_score, pos_label=1):
    fpr, tpr, thresholds = metrics.roc_curve(y_true, y_score, pos_label=1)
    roc_auc = metrics.auc(fpr, tpr)
    return roc_auc, thresholds, fpr, tpr



# -*- coding: UTF-8 -*-
from tensorflow.python.framework import ops
import tensorflow as tf

from collections import OrderedDict
import warnings

from skater.core.local_interpretation.dnni.gradient_relevance_scorer import BaseGradientMethod
from skater.core.local_interpretation.dnni.gradient_relevance_scorer import LRP
from skater.core.local_interpretation.dnni.gradient_relevance_scorer import IntegratedGradients
from skater.core.local_interpretation.dnni.perturbation_relevance_scorer import Occlusion
from skater.core.local_interpretation.dnni.initializer import Initializer

from skater.util.logger import build_logger
from skater.util.logger import _WARNING
from skater.util.logger import _INFO

logger = build_logger(_INFO, __name__)


@ops.RegisterGradient("DeepInterpretGrad")
def deep_interpreter_grad(op, grad):
    logger.debug("Computing gradient using DeepInterpretGrad")
    Initializer._grad_override_checkflag = 1
    if Initializer._enabled_method_class is not None \
            and issubclass(Initializer._enabled_method_class, BaseGradientMethod):
        logger.debug("Computing gradient using DeepInterpretGrad: {}".
                     format(Initializer._enabled_method_class._non_linear_grad))
        return Initializer._enabled_method_class._non_linear_grad(op, grad)
    else:
        logger.debug("Computing gradient using DeepInterpretGrad: {}".format(Initializer._original_grad))
        return Initializer._original_grad(op, grad)


class DeepInterpreter(object):
    """ :: Experimental :: The implementation is currently experimental and might change in future
    Interpreter for inferring Deep Learning Models. Given a trained NN model and an input vector X, DeepInterpreter
    is responsible for providing relevance scores w.r.t a target class to analyze most contributing features driving
    an estimator's decision for or against the respective class

    Framework supported: Tensorflow(>=1.4.0) and Keras(>=2.0.8)

    Parameters
    ----------
    graph : tensorflow.Graph instance
    session : tensorflow.Session to execute the graph(default session: tf.get_default_session())
    log_level : int (default: _WARNING)
        The log_level could be adjusted to other values as well. Check here `./skater/util/logger.py`

    References
    ----------
    .. [1] Ancona M, Ceolini E, Oztireli C, Gross M (ICLR, 2018).
           Towards better understanding of gradient-based attribution methods for Deep Neural Networks.
           https://arxiv.org/abs/1711.06104
           (https://github.com/marcoancona/DeepExplain/blob/master/deepexplain/tensorflow/methods.py)
    """
    __name__ = "DeepInterpreter"

    def __init__(self, graph=None, session=tf.get_default_session(), log_level=_WARNING):
        self.logger = build_logger(log_level, __name__)
        self.relevance_type = None
        self.use_case_str = None
        self.batch_size = None
        self.session = session
        if self.session is None:
            raise RuntimeError('Relevant session not retrieved')
        else:
            self.logger.info("Current session: {}".format(session.__dict__))
        self.graph = session.graph if graph is None else graph
        # request for the default graph
        self.graph_context = self.graph.as_default()
        self.override_context = self.graph.gradient_override_map(self._get_gradient_override_map())
        self.context_on = False
        self.__supported_relevance_type_dict = OrderedDict({
            'elrp': {'use_case_type': ['image'], 'method': LRP},
            'ig': {'use_case_type': ['image', 'txt'], 'method': IntegratedGradients},
            'occlusion': {'use_case_type': ['image'], 'method': Occlusion}
        })


    def __enter__(self):
        # Magic method for managing context with the usage of 'with' statement ( context-guard ).
        # This helps in managing an active keras/tensorflow session for resolving variable scope
        # reference: http://effbot.org/zone/python-with-statement.htm
        # Override gradient of all ops created in context
        self.graph_context.__enter__()
        self.override_context.__enter__()
        self.context_on = True
        return self


    def __exit__(self, type, value, traceback):
        # Exit a `with` block
        self.graph_context.__exit__(type, value, traceback)
        self.override_context.__exit__(type, value, traceback)
        self.context_on = False


    @staticmethod
    def _get_gradient_override_map():
        return dict((ops_item, 'DeepInterpretGrad') for ops_item in Initializer.activation_ops)


    def _validate_relevance_type(self, type_name, use_case_str):
        supported_type = self.__supported_relevance_type_dict[type_name] \
            if type_name in self.__supported_relevance_type_dict.keys() else None
        if supported_type is None:
            raise RuntimeError('Method type not found in {}'.format(list(self.__supported_relevance_type_dict.keys())))
        else:
            use_case_list = self.__supported_relevance_type_dict[type_name]['use_case_type']
            if use_case_str not in use_case_list:
                raise RuntimeError('Method to use-case map is not supported {}')
            else:
                return self.__supported_relevance_type_dict[type_name]['method']


    def explain(self, relevance_type, output_tensor, input_tensor, samples, use_case=None, **kwargs):
        """ Helps in computing the relevance scores for DNNs to understand the input and output behavior of the network.

        Parameters
        ----------
        relevance_type: str
            Currently, relevance score could be computed using e-LRP('elrp') or Integrated Gradient('ig'). Other
            algorithms are under development.

             - epsilon-LRP('eLRP'):
               Is recommended with Activation ops ('ReLU' and 'Tanh'). Current implementation of
               LRP works only for images and makes use of epsilon(default: 0.0001) as a stabilizer.

             - Integrated Gradient('ig'):
               Is recommended with Activation ops ('Relu', 'Elu', 'Softplus', 'Tanh', 'Sigmoid').
               It works for images and text. Optional parameters include steps(default: 100) and
               baseline(default: {'image': 'a black image'}; {'txt': zero input embedding vector})
               Gradient is computed by varying the input from the baseline(x') to the provided input(x). x, x'
               are element of R with n dimension ---> [0,1]
        output_tensor: tensorflow.python.framework.ops.Tensor
            Specify the output layer to start from
        input_tensor: tensorflow.python.framework.ops.Tensor
            Specify the input layer to reach to
        samples: numpy.array
            Batch of input for which explanations are desired.
            Note: The first dimension of the array specifies the batch size. For e.g.,
                  - for an image input of batch size 2: (2, 150, 150, 3) <batch_size, image_width, image_height, no_of_channels>
                  - for a text input of batch size 1: (1, 80) <batch_size, embedding_dimensions>
        use_case: str
            Options: 'image' or 'txt
        kwargs: optional

        Returns
        -------
        result: numpy.ndarray
            Computed relevance(contribution) score for the given input

        References
        ----------
        .. [1] Bach S, Binder A, Montavon G, Klauschen F, Müller K-R, Samek W (2015)
           On Pixel-Wise Explanations for Non-Linear Classifier Decisions by Layer-Wise Relevance Propagation.
           PLoS ONE 10(7): e0130140. https://doi.org/10.1371/journal.pone.0130140
        .. [2] Sundararajan, M, Taly, A, Yan, Q (ICML, 2017).
           Axiomatic Attribution for Deep Networks. http://arxiv.org/abs/1703.01365
        .. [3] Ancona M, Ceolini E, Öztireli C, Gross M (ICLR, 2018).
           Towards better understanding of gradient-based attribution methods for Deep Neural Networks.
           https://arxiv.org/abs/1711.06104

        Examples
        --------
        >>> from skater.core.local_interpretation.dnni.deep_interpreter import DeepInterpreter
        >>> ...
        >>> import keras
        >>> from keras.datasets import mnist
        >>>from keras.models import Sequential, Model, load_model, model_from_yaml
        >>> from keras.layers import Dense, Dropout, Flatten, Activation
        >>> from keras.layers import Conv2D, MaxPooling2D
        >>> from keras import backend as K
        >>> import tensorflow as tf
        >>> import matplotlib.pyplot as plt
        >>> sess = tf.Session()
        >>> K.set_session(sess)
        >>> ... # Load dataset
        >>> # A simple network for MNIST data-set using Keras
        >>> model = Sequential()
        >>> model.add(Conv2D(32, kernel_size=(3, 3),
        >>>         activation='relu',
        >>>         input_shape=input_shape))
        >>> model.add(Conv2D(64, (3, 3), activation='relu'))
        >>> model.add(MaxPooling2D(pool_size=(2, 2)))
        >>> model.add(Dropout(0.25))
        >>> model.add(Flatten())
        >>> model.add(Dense(128, activation='relu'))
        >>> model.add(Dropout(0.5))
        >>> model.add(Dense(num_classes))
        >>> model.add(Activation('softmax'))
        >>> ... # Compile and train the model
        >>> K.set_learning_phase(0)
        >>> with DeepInterpreter(session=K.get_session()) as di:
        >>>   # 1. Load the persisted model
        >>>   # 2. Retrieve the input tensor from the loaded model
        >>>   yaml_file = open('model_sample.yaml', 'r')
        >>>   loaded_model_yaml = yaml_file.read()
        >>>   yaml_file.close()
        >>>   loaded_model = model_from_yaml(loaded_model_yaml)
        >>>   # load weights into new model
        >>>   loaded_model.load_weights("model_mnist_cnn_3.h5")
        >>>   print("Loaded model from disk")
        >>>   input_tensor = loaded_model.layers[0].input
        >>>   output_tensor = loaded_model.layers[-2].output

        >>>    # 3. We will using the last dense layer(pre-softmax) as the output layer
        >>>    # 4. Instantiate a model with the new input and output tensor
        >>>    new_model = Model(inputs=input_tensor, outputs=output_tensor)
        >>>    target_tensor = new_model(input_tensor)
        >>>    xs = input_x
        >>>    ys = input_y
        >>>    print("X shape: {}".format(xs.shape))
        >>>    print("Y shape: {}".format(ys.shape))
        >>>    # Original Predictions
        >>>    print(loaded_model.predict_classes(xs))
        >>>    relevance_scores = di.explain('elrp', output_tensor=target_tensor * ys, input_tensor=input_tensor,
        >>>                                                                                samples=xs, use_case='image')
        """
        if not self.context_on:
            raise RuntimeError('explain can be invoked only within a DeepInterpreter context.')

        self.logger.info("all supported relevancy scorers {}".format(self.__supported_relevance_type_dict))

        # Validate if the specified relevance type is supported.
        self.relevance_type = relevance_type
        self.use_case_str = use_case
        relevance_type_class = self._validate_relevance_type(self.relevance_type, self.use_case_str)
        if relevance_type_class is None:
            raise RuntimeError('Method type not found in {}'.format(list(self.__supported_relevance_type_dict.keys())))
        self.logger.info('DeepInterpreter: executing relevance type class {}'.format(relevance_type_class))

        Initializer._grad_override_checkflag = 0
        Initializer._enabled_method_class = relevance_type_class

        method = Initializer._enabled_method_class(output_tensor, input_tensor, samples, self.session, **kwargs)
        self.logger.info('DeepInterpreter: executing method {}'.format(method))

        result = method._run()
        if issubclass(Initializer._enabled_method_class, BaseGradientMethod) and Initializer._grad_override_checkflag == 0:
            warnings.warn('Results may not reliable: As default gradient seems to have been used. '
                          'or you might have forgotten to create the graph within the DeepInterpreter context. '
                          'Be careful...')

        Initializer._enabled_method_class = None
        Initializer._grad_override_checkflag = 0
        return result



import unittest


class TestValidation(unittest.TestCase):
    pass


if __name__ == '__main__':
    suite = unittest.TestLoader().loadTestsFromTestCase(TestValidation)
    unittest.TextTestRunner(verbosity=2).run(suite)



# coding=utf-8
from skater.util import exceptions

import numbers
import numpy as np
import pandas as pd
import rpy2.robjects as ro
from rpy2.robjects.packages import importr
from rpy2.robjects import pandas2ri
pandas2ri.activate()


class BRLC(object):
    """ :: Experimental :: The implementation is currently experimental and might change in future

    BRLC(Bayesian Rule List Classifier) is a python wrapper for SBRL(Scalable Bayesian Rule list).
    SBRL is a scalable Bayesian Rule List. It's a generative estimator to build hierarchical interpretable
    decision lists. This python wrapper is an extension to the work done by Professor Cynthia Rudin,
    Benjamin Letham, Hongyu Yang, Margo Seltzer and others. For more information check out the reference section below.

    Parameters
    ----------
    iterations : int (default=30000)
        number of iterations for each MCMC chain.

    pos_sign : int (default=1)
        sign for the positive labels in the "label" column.

    neg_sign : int (default=0)
        sign for the negative labels in the "label" column.

    min_rule_len : int (default=1)
        minimum number of cardinality for rules to be mined from the data-frame.

    max_rule_len : int (default=8)
        maximum number of cardinality for rules to be mined from the data-frame.

    min_support_pos : float (default=0.1)
        a number between 0 and 1, for the minimum percentage support for the positive observations.

    min_support_neg : float (default 0.1)
        a number between 0 and 1, for the minimum percentage support for the negative observations.

    eta : int (default=1)
        a hyper-parameter for the expected cardinality of the rules in the optimal rule list.

    n_chains : int (default=10)
        number of chains

    alpha : int (default=1)
        a prior pseudo-count for the positive(alpha1) and negative(alpha0) classes. \
        Default values (1, 1)

    lambda_ : int (default=8)
        a hyper-parameter for the expected length of the rule list.

    discretize : bool (default=True)
        apply discretizer to handle continuous features.

    drop_features : bool (default=False)
        once continuous features are discretized, use this flag to either retain or drop them from the dataframe


    References
    ----------

    .. [1] Letham et.al(2015) Interpretable classifiers using rules and Bayesian analysis:
               Building a better stroke prediction model (https://arxiv.org/abs/1511.01644)

    .. [2] Yang et.al(2016) Scalable Bayesian Rule Lists (https://arxiv.org/abs/1602.08610)

    .. [3] https://github.com/Hongyuy/sbrl-python-wrapper/blob/master/sbrl/C_sbrl.py


    Examples
    --------
    >>> from skater.core.global_interpretation.interpretable_models.brlc import BRLC
    >>> import pandas as pd
    >>> from sklearn.datasets.mldata import fetch_mldata
    >>> input_df = fetch_mldata("diabetes")
    ...
    >>> Xtrain, Xtest, ytrain, ytest = train_test_split(input_df, y, test_size=0.20, random_state=0)
    >>> sbrl_model = BRLC(min_rule_len=1, max_rule_len=10, iterations=10000, n_chains=20, drop_features=True)
    >>> # Train a model, by default discretizer is enabled. So, you wish to exclude features then exclude them using
    >>> # the undiscretize_feature_list parameter
    >>> model = sbrl_model.fit(Xtrain, ytrain, bin_labels="default")
    >>> #print the learned model
    >>> sbrl_inst.print_model()
    >>> features_to_descritize = Xtrain.columns
    >>> Xtrain_filtered = sbrl_model.discretizer(Xtrain, features_to_descritize, labels_for_bin="default")
    >>> predict_scores = sbrl_model.predict_proba(Xtest)
    >>> _, y_hat  = sbrl_model.predict(Xtest)
    >>> # save and reload the model and continue with evaluation
    >>> sbrl_model.save_model("model.pkl")
    >>> sbrl_model.load_model("model.pkl")
    >>> # to access all the learned rules
    >>> sbrl_model.access_learned_rules("all")
    # For a complete example refer to rule_lists_continuous_features.ipynb or rule_lists_titanic_dataset.ipynb notebook
    """
    _estimator_type = "classifier"

    def __init__(self, iterations=30000, pos_sign=1, neg_sign=0, min_rule_len=1,
                 max_rule_len=8, min_support_pos=0.10, min_support_neg=0.10,
                 eta=1.0, n_chains=10, alpha=1, lambda_=10, discretize=True, drop_features=False):

        self.__r_sbrl = importr('sbrl')
        self.model = None
        self.__as_factor = ro.r['as.factor']
        self.__s_apply = ro.r['lapply']
        self.__r_frame = ro.r['data.frame']
        self.model_params = {
            "iters": iterations, "pos_sign": pos_sign, "neg_sign": neg_sign, "rule_minlen": min_rule_len,
            "rule_maxlen": max_rule_len, "minsupport_pos": min_support_pos, "minsupport_neg": min_support_neg,
            "eta": eta, "nchain": n_chains, "lambda": lambda_, "alpha": alpha
        }
        self.__discretize = discretize
        self.__drop_features = drop_features
        self.discretized_features = []
        self.feature_names = []


    def set_params(self, params):
        """ Set model hyper-parameters
        """
        self.model_params[list(params.keys())[0]] = list(params.values())[0]


    def discretizer(self, X, column_list, no_of_quantiles=None, labels_for_bin=None, precision=3):
        """ A discretizer for continuous features

        Parameters
        -----------
        X : pandas.DataFrame
            Dataframe containing continuous features

        column_list : list/tuple

        no_of_quantiles : int or list
            Number of quantiles, e.g. deciles(10), quartiles(4) or as a list of quantiles[0, .25, .5, .75, 1.]
            if 'None' then [0, .25, .5, .75, 1.] is used

        labels_for_bin : labels for the resulting bins

        precision : int
            precision for storing and creating bins


        Returns
        --------
        new_X: pandas.DataFrame
            Contains discretized features


        Examples
        ---------
        >>> sbrl_model = BRLC(min_rule_len=1, max_rule_len=10, iterations=10000, n_chains=20, drop_features=True)
        >>> ...
        >>> features_to_descritize = Xtrain.columns
        >>> Xtrain_discretized = sbrl_model.discretizer(Xtrain, features_to_descritize, labels_for_bin="default")
        >>> predict_scores = sbrl_model.predict_proba(Xtrain_discretized)
        """
        if not isinstance(X, pd.DataFrame):
            raise TypeError("Only pandas.DataFrame as input type is currently supported")
        q_value = [0, .25, .5, .75, 1.] if no_of_quantiles is None else no_of_quantiles
        q_labels = [1, 2, 3, 4] if labels_for_bin is 'default' else labels_for_bin
        new_X = X.copy()
        for column_name in column_list:
            new_clm_name = '{}_q_label'.format(column_name)
            self.discretized_features.append(new_clm_name)
            new_X.loc[:, new_clm_name] = pd.qcut(X[column_name].rank(method='first'), q=q_value,
                                                 labels=q_labels, duplicates='drop', precision=precision)

            # Drop the continuous feature column which has been discritized
            new_X = new_X.drop([column_name], axis=1) if self.__drop_features else new_X
            # explicitly convert the labels column to 'str' type
            new_X = new_X.astype(dtype={'{}_q_label'.format(column_name): "str"})
        return new_X


    def _filter_continuous_features(self, X, column_list=None):
        import collections
        # Sequence is a base class for list and tuple. column_list could be of either type
        if not isinstance(column_list, collections.Sequence):
            raise TypeError("Only list/tuple type supported for specifying column list")
        c_l = X.columns if column_list is None else column_list
        # To check for numeric type, validate again numbers.Number (base class for numeric type )
        # Reference[PEP-3141]: https://www.python.org/dev/peps/pep-3141/
        numeric_type_columns = tuple(filter(lambda c_name: isinstance(X[c_name].iloc[0], numbers.Number), c_l))
        return numeric_type_columns


    # a helper function to filter unwanted features
    filter_to_be_discretize = lambda self, clmn_list, unwanted_list: \
        tuple(filter(lambda c_name: c_name not in unwanted_list, clmn_list))


    def fit(self, X, y_true, n_quantiles=None, bin_labels='default', undiscretize_feature_list=None, precision=3):
        """ Fit the estimator.

        Parameters
        -----------
            X : pandas.DataFrame object, that could be used by the model for training.
                 It must not have a column named 'label'

            y_true : pandas.Series, 1-D array to store ground truth labels

        Returns
        -------
            SBRL model instance: rpy2.robjects.vectors.ListVector


        Examples
        ---------
        >>> from skater.core.global_interpretation.interpretable_models.brlc import BRLC
        >>> sbrl_model = BRLC(min_rule_len=1, max_rule_len=10, iterations=10000, n_chains=20, drop_features=True)
        >>> # Train a model, by default discretizer is enabled. So, you wish to exclude features then exclude them using
        >>> # the undiscretize_feature_list parameter
        >>> model = sbrl_model.fit(Xtrain, ytrain, bin_labels="default")
        """
        if len(np.unique(y_true)) != 2:
            raise Exception("Supports only binary classification right now")

        if not isinstance(X, pd.DataFrame):
            raise exceptions.DataSetError("Only pandas.DataFrame as input type is currently supported")

        # Conditions being managed
        # 1. if 'undiscretize_feature_list' is empty and discretization flag is enabled,
        #    discretize 'all' continuous features
        # 2. if undiscretize_feature_list is not empty and discretization flag is enabled, filter the ones not needed
        #    needed
        for_discretization_clmns = tuple(filter(lambda c_name: c_name not in undiscretize_feature_list, X.columns)) \
            if undiscretize_feature_list is not None else tuple(X.columns)

        data = self.discretizer(X, self._filter_continuous_features(X, for_discretization_clmns),
                                no_of_quantiles=n_quantiles, labels_for_bin=bin_labels, precision=precision) \
            if self.__discretize is True else X

        # record all the feature names
        self.feature_names = data.columns
        data.loc[:, "label"] = y_true
        data_as_r_frame = self.__r_frame(self.__s_apply(data, self.__as_factor))
        self.model = self.__r_sbrl.sbrl(data_as_r_frame, **self.model_params)
        return self.model


    def save_model(self, model_name, compress=True):
        """ Persist the model for future use
        """
        import joblib
        if self.model is not None:
            joblib.dump(self.model, model_name, compress=compress)
        else:
            raise Exception("SBRL model is not fitted yet; no relevant model instance present")


    def load_model(self, serialized_model_name):
        """ Load a serialized model
        """
        import joblib
        try:
            self.model = joblib.load(serialized_model_name)
            # update the BRLC model instance with the the uploaded model
        except (OSError, IOError) as err:
            print("Something is not right with the serialization format. Details {}".format(err))
            raise


    def predict_proba(self, X):
        """ Computes possible class probabilities for the input 'X'

        Parameters
        -----------
            X: pandas.DataFrame object

        Returns
        -------
            pandas.DataFrame of shape (#datapoints, 2), the possible probability of each class for each observation
        """
        if not isinstance(X, pd.DataFrame):
            raise exceptions.DataSetError("Only pandas.DataFrame as input type is currently supported")

        data_as_r_frame = self.__r_frame(self.__s_apply(X, self.__as_factor))
        results = self.__r_sbrl.predict_sbrl(self.model, data_as_r_frame)
        return pandas2ri.ri2py_dataframe(results).T


    def predict(self, X=None, prob_score=None, threshold=0.5, pos_label=1):
        """ Predict the class for input 'X'
        The predicted class is determined by setting a threshold. Adjust threshold to
        balance between sensitivity and specificity

        Parameters
        -----------
        X: pandas.DataFrame
            input examples to be scored
        prob_score: pandas.DataFrame or None (default=None)
            If set to None, `predict_proba` is called before computing the class labels.
            If you have access to probability scores already, use the dataframe of probability scores to compute the
            final class label
        threshold: float (default=0.5)
        pos_label: int (default=1)
            specify how to identify positive label

        Returns
        -------
        y_prob, y_prob['label]: pandas.Series, numpy.ndarray
            Contains the probability score for the input 'X'

        """
        # TODO: Extend it for multi-class classification
        probability_df = self.predict_proba(X) if X is not None and prob_score is None else prob_score
        y_prob = probability_df.loc[:, pos_label]
        y_prob['label'] = np.where(y_prob.values > threshold, 1, 0)
        return y_prob, y_prob['label']


    def print_model(self):
        """ print the decision stumps of the learned estimator
        """
        self.__r_sbrl.print_sbrl(self.model)


    def access_learned_rules(self, rule_indexes="all"):
        """ Access all learned decision rules. This is useful for building and developing intuition

        Parameters
        ----------
        rule_indexes: str (default="all", retrieves all the rules)
            Specify the index of the rules to be retrieved
            index could be set as 'all' or a range could be specified e.g. '(1:3)' will retrieve the rules 1 and 2
        """
        if not isinstance(rule_indexes, str):
            raise TypeError('Expected type string {} provided'.format(type(rule_indexes)))

        # Convert model properties into a readable python dict
        result_dict = dict(zip(self.model.names, map(list, list(self.model))))

        indexes_func = lambda indexes: [int(v) for v in indexes.split(':')]
        # original index starts from 0 while the printed index starts from 1, hence adjust the index
        rules_filter = lambda all_rules, indexes: all_rules['rulenames'][(indexes[0] - 1):(indexes[1] - 1)] \
            if rule_indexes.find(':') > -1 else all_rules['rulenames'][indexes[0] - 1]

        # Enable the ability to access single or multiple sequential model learned decisions
        rules_result = result_dict['rulenames'] if rule_indexes == "all" \
            else rules_filter(result_dict, indexes_func(rule_indexes))
        return rules_result



import unittest

import keras
from keras.datasets import mnist
from keras.models import Sequential, Model, load_model, model_from_yaml
from keras import backend as K

from skater.core.local_interpretation.dnni.deep_interpreter import DeepInterpreter


class TestDNNI(unittest.TestCase):

    # The below architecture maps to the pre_trained_models/mnist_cnn/model_mnist_cnn_epoch_3*
    # @classmethod
    # def _build_mnist_CNN_model(cls):
    #     from keras.layers import Dense, Dropout, Flatten, Activation
    #     from keras.layers import Conv2D, MaxPooling2D
    #
    #     Build and train a network.
    #     model = Sequential()
    #     model.add(Conv2D(32, kernel_size=(3, 3), activation='relu', input_shape=cls.input_shape))
    #     model.add(Conv2D(64, (3, 3), activation='relu'))
    #     model.add(MaxPooling2D(pool_size=(2, 2)))
    #     model.add(Dropout(0.25))
    #     model.add(Flatten())
    #     model.add(Dense(128, activation='relu'))
    #     model.add(Dropout(0.5))
    #     model.add(Dense(cls.num_classes))
    #     model.add(Activation('softmax'))
    #
    #     model.compile(loss=keras.losses.categorical_crossentropy, optimizer=keras.optimizers.Adadelta(),
    #                   metrics=['accuracy'])
    #     model.fit(cls.x_train, cls.y_train, batch_size=cls.batch_size, epochs=cls.epochs,
    #               verbose=1, validation_data=(cls.x_test, cls.y_test))
    #     return model


    @classmethod
    def _load(cls, model_name):
        f_h = open('{}.yaml'.format(model_name), 'r')
        persisted_yaml_model = f_h.read()
        f_h.close()
        loaded_model = model_from_yaml(persisted_yaml_model)
        # load weights into retrieved model instance
        loaded_model.load_weights("{}.h5".format(model_name))
        return loaded_model


    @classmethod
    def setUpClass(cls):
        # MNIST dataset used for building pre_trained_models/mnist_cnn/model_mnist_cnn_epoch_3
        cls.batch_size = 128
        cls.num_classes = 10
        cls.epochs = 2
        # input image dimensions
        cls.img_rows, cls.img_cols = 28, 28
        # shuffled and split between train and test sets
        (cls.x_train, cls.y_train), (cls.x_test, cls.y_test) = mnist.load_data()
        if K.image_data_format() == 'channels_first':
            cls.x_train = cls.x_train.reshape(cls.x_train.shape[0], 1, cls.img_rows, cls.img_cols)
            cls.x_test = cls.x_test.reshape(cls.x_test.shape[0], 1, cls.img_rows, cls.img_cols)
            cls.input_shape = (1, cls.img_rows, cls.img_cols)
        else:
            cls.x_train = cls.x_train.reshape(cls.x_train.shape[0], cls.img_rows, cls.img_cols, 1)
            cls.x_test = cls.x_test.reshape(cls.x_test.shape[0], cls.img_rows, cls.img_cols, 1)
            cls.input_shape = (cls.img_rows, cls.img_cols, 1)

        cls.x_train = cls.x_train.astype('float32')
        cls.x_test = cls.x_test.astype('float32')
        cls.x_train /= 255
        cls.x_test /= 255
        cls.x_train = (cls.x_train - 0.5) * 2
        cls.x_test = (cls.x_test - 0.5) * 2

        # convert class vectors to binary class matrices
        cls.y_train = keras.utils.to_categorical(cls.y_train, cls.num_classes)
        cls.y_test = keras.utils.to_categorical(cls.y_test, cls.num_classes)


    def test_deep_interpreter_cnn(self):
        K.set_learning_phase(0)
        with DeepInterpreter(session=K.get_session()) as di:
            # 1. Load the persisted model
            # 2. Retrieve the input tensor from the loaded model

            retrieved_model = self._load('skater/tests/pre_trained_models/mnist_cnn/model_mnist_cnn_epoch_3')
            input_tensor = retrieved_model.layers[0].input
            output_tensor = retrieved_model.layers[-2].output

            # 3. We will using the last dense layer(pre-softmax) as the output layer
            # 4. Instantiate a model with the new input and output tensor
            new_model = Model(inputs=input_tensor, outputs=output_tensor)
            target_tensor = new_model(input_tensor)
            xs = self.x_test[0:2]
            ys = self.y_test[0:2]

            relevance_scores_elrp = di.explain('elrp', target_tensor * ys, input_tensor, xs, use_case='image')
            relevance_scores_ig = di.explain('ig', target_tensor * ys, input_tensor, xs, use_case='image')
            # test for optional arguments related to window_size and step size for 'occlusion'
            optional_args = {"window_size": 2, "step": 2}
            relevance_scores_occ = di.explain('occlusion', target_tensor * ys, input_tensor, xs,
                                              use_case='image', **optional_args)
        self.assertEquals(relevance_scores_elrp.shape, (2, 28, 28, 1))
        self.assertEquals(relevance_scores_ig.shape, (2, 28, 28, 1))
        self.assertEquals(relevance_scores_occ.shape, (2, 28, 28, 1))


if __name__ == '__main__':
    suite = unittest.TestLoader().loadTestsFromTestCase(TestDNNI)
    unittest.TextTestRunner(verbosity=2).run(suite)



from bs4 import BeautifulSoup
import re
import numpy as np


def preprocessor(text):
    # remove HTML tags
    text = BeautifulSoup(text, 'html.parser').get_text()

    # regex for matching emoticons, keep emoticons, ex: :), :-P, :-D
    r = '(?::|;|=|X)(?:-)?(?:\)|\(|D|P)'
    emoticons = re.findall(r, text)
    text = re.sub(r, '', text)

    # convert to lowercase and append all emoticons behind (with space in between)
    # replace('-','') removes nose of emoticons
    text = re.sub('[\W]+', ' ', text.lower()) + ' ' + ' '.join(emoticons).replace('-', '')
    return text


def cleaner(text, to_lower=True, norm_num=False, char_to_strip=' |(|)|,', non_alphanumeric_exceptions=","):
    # if the to_lower flag is true, convert the text to lowercase
    text = text.lower() if to_lower else text
    # Removes unwanted http hyper links in text
    text = re.sub(r"http(s)?://\S*", " ", text)
    # In some cases, one may want to normalize numbers for better visualization
    text = re.sub(r"[0-9]", "1", text) if norm_num else text
    # remove non-alpha numeric characters [!, $, #, or %] and normalize whitespace
    text = re.sub(r"[^A-Za-z0-9-" + non_alphanumeric_exceptions + "]", " ", text)
    # replace leftover unwanted white space
    text = whitespace_normalizer(text)
    # remove trailing or leading white spaces
    text = text.strip(char_to_strip)
    return text


# returns indexes where ground truth and predicted value does not match
query_for_false_predictions = lambda predictions, ground_truth: np.where(ground_truth != predictions)

# whitespace normalizer
whitespace_normalizer = lambda x: re.sub(r"\s+", " ", x)

# Convert string to a words list
generate_word_list = lambda x, token_type: whitespace_normalizer(x).split(token_type)



# -*- coding: UTF-8 -*-

from itertools import cycle
import numpy as np
import pandas as pd
from functools import partial
from multiprocess import Pool

from ...data import DataManager
from .base import BaseGlobalInterpretation
from ...util.plotting import COLORS
from ...util.exceptions import *
from ...util.dataops import divide_zerosafe
from ...util.progressbar import ProgressBar
from ...util.static_types import StaticTypes


class FeatureImportance(BaseGlobalInterpretation):
    """Contains methods for feature importance. Subclass of BaseGlobalInterpretation.

    """

    def feature_importance(self, model_instance, ascending=True, filter_classes=None, n_jobs=-1,
                           progressbar=True, n_samples=5000, method='prediction-variance', scorer_type='default',
                           use_scaling=False):

        """
        Computes feature importance of all features related to a model instance.
        Supports classification, multi-class classification, and regression.


        Parameters
        ----------
        model_instance: skater.model.model.Model subtype
            the machine learning model "prediction" function to explain, such that
            predictions = predict_fn(data).
        ascending: boolean, default True
            Helps with ordering Ascending vs Descending
        filter_classes: array type
            The classes to run partial dependence on. Default None invokes all classes.
            Only used in classification models.
        n_jobs: int
            How many concurrent processes to use. Defaults -1, which grabs as many as are available.
            Use 1 to avoid multiprocessing altogether.
        progressbar: bool
            Whether to display progress. This affects which function we use to operate on the pool
            of processes, where including the progress bar results in 10-20% slowdowns.
        n_samples: int
            How many samples to use when computing importance.
        method: string (default 'prediction-variance'; 'model-scoring' for estimator specific scoring metric
            How to compute feature importance. 'model-scoring' requires Interpretation.training_labels.
            Note this choice should only rarely makes any significant differences
            prediction-variance: mean absolute value of changes in predictions, given perturbations.
            model-scoring: difference in log_loss or MAE of training_labels given perturbations.
        scorer_type: string
            only used when method='model-scoring', and in this case defines which scoring function to use.
            Default value is 'default', which evaluates to:
                regressors: mean absolute error
                classifiers with probabilities: cross entropy
                classifiers without probabilities: f1 score
            See Skater.model.scorers for details.
        use_scaling: bool
            Whether to weight the importance values by the strength of the perturbations.
            Generally doesn't effect results unless n_samples is very small.


        Returns
        -------
        importances : Sorted Series


        References
        ----------
        Wei, Pengfei, Zhenzhou Lu, and Jingwen Song. "Variable Importance Analysis: A Comprehensive Review".
        Reliability Engineering & System Safety 142 (2015): 399-432.

        Examples
        --------

            >>> from skater.model import InMemoryModel
            >>> from skater.core.explanations import Interpretation
            >>> from sklearn.ensemble import RandomForestClassifier
            >>> rf = RandomForestClassifier()
            >>> rf.fit(X,y)
            >>> model = InMemoryModel(rf.predict_proba, examples = X)
            >>> interpreter = Interpretation()
            >>> interpreter.load_data(X)
            >>> interpreter.feature_importance.feature_importance(model)


        """

        if filter_classes:
            err_msg = "members of filter classes must be" \
                      "members of model_instance.classes." \
                      "Expected members of: {0}\n" \
                      "got: {1}".format(model_instance.target_names,
                                        filter_classes)
            filter_classes = list(filter_classes)
            assert all([i in model_instance.target_names for i in filter_classes]), err_msg

        if n_samples <= self.data_set.n_rows:
            inputs, labels = self.data_set.generate_sample(strategy='random-choice',
                                                           include_y=True,
                                                           sample=True,
                                                           n_samples=n_samples)
        else:
            inputs, labels = self.data_set.X, self.data_set.y

        if method == 'model-scoring' and labels is None:
            raise FeatureImportanceError("If labels are not set, you"
                                         "can only use feature importance methods that do "
                                         "not require ground truth labels")

        original_predictions = model_instance.predict(inputs)
        n_jobs = None if n_jobs < 0 else n_jobs
        predict_fn = model_instance._get_static_predictor()
        executor_instance = Pool(n_jobs)
        arg_list = self.data_set.feature_ids
        scorer = model_instance.scorers.get_scorer_function(scorer_type=scorer_type)

        if progressbar:
            self.interpreter.logger.warn("Progress bars slow down runs by 10-20%. For slightly \n"
                                         "faster runs, do progress_bar=False")
            n_iter = len(self.data_set.feature_ids)
            p = ProgressBar(n_iter, units='features')
            mapper = executor_instance.imap
        else:
            mapper = executor_instance.map

        fi_func = partial(compute_feature_importance,
                          input_data=inputs,
                          estimator_fn=predict_fn,
                          original_predictions=original_predictions,
                          feature_info=self.data_set.feature_info,
                          feature_names=self.data_set.feature_ids,
                          training_labels=labels,
                          method=method,
                          scaled=use_scaling,
                          scorer=scorer)

        importances = {}
        try:
            if n_jobs == 1:
                raise ValueError("Skipping to single processing")
            importance_dicts = []
            for importance in mapper(fi_func, arg_list):
                importance_dicts.append(importance)
                if progressbar:
                    p.animate()
        except:
            self.interpreter.logger.warn("Multiprocessing failed, going single process")
            importance_dicts = []
            for importance in map(fi_func, arg_list):
                importance_dicts.append(importance)
                if progressbar:
                    p.animate()
        finally:
            executor_instance.close()
            executor_instance.join()
            executor_instance.terminate()

        for i in importance_dicts:
            importances.update(i)

        importances = pd.Series(importances).sort_values(ascending=ascending)

        if not importances.sum() > 0:
            self.interpreter.logger.debug("Importances that caused a bug: {}".format(importances))
            raise(FeatureImportanceError("Something went wrong. Importances do not sum to a positive value\n"
                                         "This could be due to:\n"
                                         "1) 0 or infinite divisions\n"
                                         "2) perturbed values == original values\n"
                                         "3) feature is a constant\n"
                                         "Please submit an issue here:\n"
                                         "https://github.com/datascienceinc/Skater/issues"))

        importances = divide_zerosafe(importances, (np.ones(importances.shape[0]) * importances.sum()))
        return importances


    def plot_feature_importance(self, modelinstance, filter_classes=None, ascending=True, ax=None, progressbar=True,
                                n_jobs=-1, n_samples=5000, method='prediction-variance', scorer_type='default',
                                use_scaling=False):

        """Computes feature importance of all features related to a model instance,
        then plots the results. Supports classification, multi-class classification, and regression.

        Parameters
        ----------
        modelinstance: skater.model.model.Model subtype
            estimator "prediction" function to explain the predictive model. Could be probability scores
            or target values
        filter_classes: array type
            The classes to run partial dependence on. Default None invokes all classes.
            Only used in classification models.
        ascending: boolean, default True
            Helps with ordering Ascending vs Descending
        ax: matplotlib.axes._subplots.AxesSubplot
            existing subplot on which to plot feature importance. If none is provided,
            one will be created.
        progressbar: bool
            Whether to display progress. This affects which function we use to operate on the pool
            of processes, where including the progress bar results in 10-20% slowdowns.
        n_jobs: int
            How many concurrent processes to use. Defaults -1, which grabs as many as are available.
            Use 1 to avoid multiprocessing altogether.
        n_samples: int
            How many samples to use when computing importance.
        method: string
            How to compute feature importance. 'model-scoring' requires Interpretation.training_labels
            prediction-variance: mean absolute value of changes in predictions, given perturbations.
            model-scoring: difference in log_loss or MAE of training_labels given perturbations.
            Note this vary rarely makes any significant differences
        scorer_type: string
            only used when method='model-scoring', and in this case defines which scoring function to use.
            Default value is 'default', which evaluates to:
                regressors: mean absolute error
                classifiers with probabilities: cross entropy
                classifiers without probabilities: f1 score
            See Skater.model.scorers for details.
        use_scaling: bool
            Whether to weight the importance values by the strength of the perturbations.
            Generally doesn't effect results unless n_samples is very small.


        Returns
        -------
        f: figure instance
        ax: matplotlib.axes._subplots.AxesSubplot
            could be used to for further modification to the plots

        Examples
        --------
            >>> from skater.model import InMemoryModel
            >>> from skater.core.explanations import Interpretation
            >>> from sklearn.ensemble import RandomForestClassifier
            >>> rf = RandomForestClassifier()
            >>> rf.fit(X,y)
            >>> model = InMemoryModel(rf.predict_proba, examples = X)
            >>> interpreter = Interpretation()
            >>> interpreter.load_data(X)
            >>> interpreter.feature_importance.plot_feature_importance(model, ascending=True, ax=ax)
            """
        try:
            global pyplot
            from matplotlib import pyplot
        except ImportError:
            raise (MatplotlibUnavailableError("Matplotlib is required but unavailable on your system."))
        except RuntimeError:
            raise (MatplotlibDisplayError("Matplotlib unable to open display"))

        importances = self.feature_importance(modelinstance,
                                              filter_classes=filter_classes,
                                              progressbar=progressbar,
                                              n_samples=n_samples,
                                              n_jobs=n_jobs,
                                              method=method,
                                              scorer_type=scorer_type,
                                              use_scaling=use_scaling)

        if ax is None:
            f, ax = pyplot.subplots(1)
        else:
            f = ax.figure

        colors = cycle(COLORS)
        color = next(colors)
        importances.sort_values(ascending=ascending).plot(kind='barh', ax=ax, color=color)
        return f, ax


def compute_feature_importance(feature_id, input_data, estimator_fn,
                               original_predictions, feature_info,
                               feature_names, training_labels=None,
                               method='prediction-variance',
                               scaled=False, scorer=None):
    """Global function for computing column-wise importance

    Parameters
    ----------
    feature_id: hashable
        which feature for which to compute importance
    input_data:
        training data (or sample)
    estimator_fn: callable
        prediction function
    original_predictions: array type
        predicted values before perturbation
    feature_info: dict
        from skater.data.DataManager.feature_info
    feature_names: array type
        list of feature names
    training_labels: array type
        ground truth labels. only required if method="perfomance decrease"
    method: string
        prediction-variance: importance based on changes in predictions caused by
                             changing values of the feature space.
        model-scoring:       importance based on differences in prediction scores/cost
                             functions caused by changing values of the feature space.
    model_type: string
        regression or classification
    scaled: bool
        Whether to weight the importance values by the stregth of the perturbations.

    Returns
    ----------
    importance: dict
        {feature id: importance value}
    """

    copy_of_data_set = DataManager(input_data.copy(), feature_names=feature_names)
    n = copy_of_data_set.n_rows

    original_values = copy_of_data_set[feature_id]

    # collect perturbations
    if feature_info[feature_id]['numeric']:
        samples = copy_of_data_set.generate_column_sample(feature_id, n_samples=n,
                                                          strategy='uniform-over-similarity-ranks')
    else:
        samples = copy_of_data_set.generate_column_sample(feature_id, n_samples=n, strategy='random-choice')


    copy_of_data_set[feature_id] = samples.reshape(-1)

    if copy_of_data_set.data_type == np.ndarray:
        new_predictions = estimator_fn(copy_of_data_set.values)
    else:
        new_predictions = estimator_fn(copy_of_data_set.X)

    importance = compute_importance(new_predictions,
                                    original_predictions,
                                    original_values,
                                    samples,
                                    training_labels,
                                    method=method,
                                    scaled=scaled,
                                    scorer=scorer)
    return {feature_id: importance}


def _compute_importance_via_prediction_variance(new_predictions, original_predictions,
                                                original_x, perturbed_x, scaled=True):
    """Mean absolute change in predictions given perturbations in a feature"""
    changes_in_predictions = abs(new_predictions - original_predictions)
    if len(changes_in_predictions.shape) == 1:
        changes_in_predictions = changes_in_predictions[:, np.newaxis]

    if scaled:
        scales = importance_scaler(original_x, perturbed_x)
        changes_in_predictions = np.sum(changes_in_predictions, axis=1) * scales

    importance = np.mean(changes_in_predictions)
    return importance


def _compute_importance_via_model_scoring(new_predictions, original_predictions, training_labels,
                                          original_x, perturbed_x, scorer, scaled=True):

    """Mean absolute error of predictions given perturbations in a feature"""
    if scaled:
        sample_weight = importance_scaler(original_x, perturbed_x)
    else:
        sample_weight = None

    score1 = scorer(training_labels, new_predictions, sample_weight=sample_weight)
    score2 = scorer(training_labels, original_predictions, sample_weight=sample_weight)
    multiple = 1 if scorer.type == StaticTypes.scorer_types.increasing else -1
    return abs(max((score2 - score1) * multiple, 0))


def importance_scaler(original_x, perturbed_x):
    perturbed_x = perturbed_x.reshape(original_x.shape)

    scales = abs(perturbed_x - original_x) / (max(original_x) - min(original_x))
    if sum(scales == 0):
        scales = np.ones(perturbed_x.shape[0])

    return scales


def compute_importance(new_predictions, original_predictions, original_x, perturbed_x,
                       training_labels, method='prediction-variance', scaled=False,
                       scorer=None):
    if method == 'prediction-variance':
        importance = _compute_importance_via_prediction_variance(np.array(new_predictions),
                                                                 np.array(original_predictions),
                                                                 np.array(original_x),
                                                                 np.array(perturbed_x),
                                                                 scaled)
    elif method == 'model-scoring':
        importance = _compute_importance_via_model_scoring(np.array(new_predictions),
                                                           np.array(original_predictions),
                                                           training_labels,
                                                           np.array(original_x),
                                                           np.array(perturbed_x),
                                                           scorer,
                                                           scaled)

    else:
        raise(KeyError("Unrecongized method for computing feature_importance: {}".format(method)))
    return importance



# -*- coding: utf-8 -*-
from __future__ import print_function
import time
import sys
import os


class ProgressBar(object):
    """Progress bar

    Parameters:

        n_cycles(int):
            The number of iterations.

        units(str):
            The unit name to be display (e.g. 10 iterations...)

    """
    def __init__(self, n_cycles, units='iterations'):

        self.start_time = time.time()
        self.n_cycles = n_cycles
        self.current_cycle = 0
        self.units = units
        self.todo_s = "-"
        self.complete_s = '█'
        self.bar_length = 20


    def printer(self, s):
        sys.stdout.write("\r{}".format(s))
        sys.stdout.flush()


    def format_count(self, x):
        """
        x is between 0 and 1
        if bar_length is 20, then complete % needs to .05
        """
        return int((x * 100) / 5)


    def animate(self):
        self.current_cycle += 1
        elapsed = int(time.time() - self.start_time)
        time_elapsed = " Time elapsed: {} seconds".format(elapsed)
        counter = "[{0}/{1}] {2} ".format(self.current_cycle, self.n_cycles, self.units)

        n_complete_to_display = self.format_count(self.current_cycle / float(self.n_cycles))
        n_todo_to_display = self.bar_length - n_complete_to_display
        bar = self.complete_s * n_complete_to_display + self.todo_s * n_todo_to_display
        msg = counter + bar + time_elapsed
        self.printer(msg)


# def is_run_from_ipython():
#     try:
#         __IPYTHON__
#         return True
#     except NameError:
#         return False



# -*- coding: UTF-8 -*-
import numpy as np

from skater.core.local_interpretation.dnni.initializer import Initializer
from skater.util.logger import build_logger
from skater.util.logger import _INFO


class BasePerturbationMethod(Initializer):
    """
    Base class for perturbation-based relevance/attribution computation

    """

    __name__ = "BasePerturbationMethod"
    logger = build_logger(_INFO, __name__)

    def __init__(self, output_tensor, input_tensor, samples, current_session):
        super(BasePerturbationMethod, self).__init__(output_tensor, input_tensor, samples, current_session)


class Occlusion(BasePerturbationMethod):
    """ Occlusion is a perturbation based inference algorithm. Such forms of algorithm direcly computes the
    relevance/attribution of the input features :math:`(X_{i})` by systematically occluding different
    portions of the image (by removing, masking or altering them), then running a forward pass on the new input to
    produce a new output, and then measuring and monitoring the difference between the original output and new output.
    Perturbation based interpretation helps one to compute direct estimation of the marginal effect of a feature but
    the inference might be computationally expensive depending on the cardinatlity of the feature space.
    The choice of the baseline value while perturbing through the feature space could be set to 0,
    as explained in detail by Zeiler & Fergus, 2014[2].

    References
    ----------
    .. [1] Ancona M, Ceolini E, Oztireli C, Gross M (ICLR, 2018).
    .. Towards better understanding of gradient-based attribution methods for Deep Neural Networks.
    .. [2] Zeiler, M and Fergus, R (Springer, 2014). Visualizing and understanding convolutional networks.
    .. In European conference on computer vision, pp. 818–833.
    .. [3] https://github.com/marcoancona/DeepExplain/blob/master/deepexplain/tensorflow/methods.py
    """
    __name__ = "Occlusion"
    logger = build_logger(_INFO, __name__)

    def __init__(self, output_tensor, input_tensor, samples, current_session, **kwargs):
        super(Occlusion, self).__init__(output_tensor, input_tensor, samples, current_session)

        self.input_shape = samples.shape[1:]
        self.replace_value = kwargs['replace_value'] if 'replace_value' in kwargs.keys() else 0
        self.window_size = kwargs['window_size'] if 'window_size' in kwargs.keys() else 1
        self.step = kwargs['step'] if 'step' in kwargs.keys() else 1
        # the input samples are expected to be of the shape,
        # (1, 150, 150, 3) <batch_size, image_width, image_height, no_of_channels>
        self.batch_size = self.samples.shape[0]
        Occlusion.logger.info('Input shape: {}; window_size/step: ({}/{}); replace value: {}; batch size: {}'.
                              format(self.input_shape, self.window_size, self.step, self.replace_value, self.batch_size))


    def _create_masked_input(self, row_value, col_value):
        masked_input = np.array(self.samples)
        # mask the region as set by the window size by replacing the pixel values with the specified value(default:0)
        masked_input[:, row_value:(row_value + self.window_size), col_value:(col_value + self.window_size), :] = self.replace_value
        return masked_input


    def _run(self):
        mask = np.array([self.batch_size, self.window_size, self.window_size, self.samples[0].shape[2]])
        mask.fill(self.replace_value)
        Occlusion.logger.info('Shape of the mask patch: {}'.format(mask.shape))
        relevance_score = np.zeros_like(self.samples, dtype=np.float32)
        # normalizer matrix is set to 1 default; as matrix cell gets used atleast once
        normalizer = np.ones_like(relevance_score)

        # Compute original output
        default_eval = self._session_run(self.output_tensor, self.samples)
        Occlusion.logger.info("shape of the default eval value :{}".format(default_eval.shape))

        count = 1  # to keep track of the number of times a matrix cell is used while perturbing through the feature space
        # Perturb through the feature space by replacing and masking
        for row in range(0, self.samples[0].shape[0] - self.window_size, self.step):
            for col in range(0, self.samples[0].shape[1] - self.window_size, self.step):
                # create masked input while rolling through the input matrix
                new_input = self._create_masked_input(row, col)

                # compute entropy when compared to the original eval value
                delta = default_eval - self._session_run(self.output_tensor, new_input)
                delta_aggregated = np.sum(delta.reshape((self.batch_size, -1)), -1, keepdims=True)
                relevance_score[:, row:(row + self.window_size), col:(col + self.window_size), :] += delta_aggregated

                # keeping track of the number of time a matrix cell is used while perturbing feature space based
                # on window size
                normalizer[:, row:(row + self.window_size), col:(col + self.window_size), :] += (count - 1)

        Occlusion.logger.info("Min/Max normalizer weight: {}/{}".format(np.min(normalizer.shape),
                                                                        np.max(normalizer.shape)))
        relevance_score_norm = relevance_score / normalizer
        Occlusion.logger.info("relevance score matrix shape :{}".format(relevance_score_norm.shape))
        return relevance_score_norm



# -*- coding: UTF-8 -*-

import numpy as np
import skimage
import skimage.io
from scipy import ndimage
from skimage import util
from skimage.transform import rotate
from skimage import exposure

from .exceptions import MatplotlibUnavailableError
from .exceptions import DataSetError
from skater.util.logger import build_logger
from skater.util.logger import _INFO

__all__ = ['add_noise', 'image_transformation', 'flip_pixels', 'normalize', 'show_image', 'greater_than', 'less_than',
           'equal_to', 'greater_than_or_equal', 'less_than_equal', 'in_between', 'view_windows']


logger = build_logger(_INFO, __name__)


def load_image(path, img_height=None, img_width=None, crop_from_center=True, rgb2bgr=False, mode='constant',
               preserve_range=True, anti_aliasing=None, anti_aliasing_sigma=None):
    if not img_height or img_width is None:
        raise AssertionError("Dimensions for the image size is not specified as reqiured. Either image height or "
                             "image weight is missing ")

    # load image
    img = skimage.io.imread(path)

    # Crop image from the center
    short_edge = np.min(img.shape[:2])
    yy = int((img.shape[0] - short_edge) / 2)
    xx = int((img.shape[1] - short_edge) / 2)
    crop_img = img[yy: yy + short_edge, xx: xx + short_edge]

    in_img = crop_img if crop_from_center is True else img

    # If RGB --> BGR flag is enabled
    in_img = in_img[:, :, ::-1] if rgb2bgr else in_img
    # Re-size the image to required dimension
    resized_img = skimage.transform.resize(in_img, (img_height, img_width), mode=mode,
                                           preserve_range=preserve_range, anti_aliasing=anti_aliasing,
                                           anti_aliasing_sigma=anti_aliasing_sigma)
    return resized_img


def add_noise(image, noise_typ='gaussian', random_state=None):
    """ Helper function to add different forms of noise

    Parameters
    -----------
    image: numpy.ndarray
        input image
    noise_typ: string ( default 'gaussian' )
        - 'gaussian'  Gaussian-distributed additive noise.
        - 'localvar'  Gaussian-distributed additive noise, with specified
                      local variance at each point of `image`
        - 'poisson'   Poisson-distributed noise generated from the data.
        - 'salt'      Replaces random pixels with 1.
        - 'pepper'    Replaces random pixels with 0 (for unsigned images) or
                      -1 (for signed images).
        - 's&p'       Replaces random pixels with either 1 or `low_val`, where
                      `low_val` is 0 for unsigned images or -1 for signed
                      images.
        - 'speckle'   Multiplicative noise using out = image + n*image, where
                      n is uniform noise with specified mean & variance.
    random_state: int
        seed for setting repeatable state before generating noise

    Returns
    -------
    numpy.ndarray

    References
    ----------
    .. [1] http://scikit-image.org/docs/stable/api/skimage.util.html#random-noise
    """
    return skimage.util.random_noise(image, mode=noise_typ, seed=random_state)


def _rescale_intensity(X, q=(0.2, 99.8)):
    v_min, v_max = np.percentile(X, q)
    return exposure.rescale_intensity(X, in_range=(v_min, v_max))


def image_transformation(X, method_type='blur', **kwargs):
    # https://www.kaggle.com/tomahim/image-manipulation-augmentation-with-skimage
    q = kwargs['percentile'] if 'percentile' in kwargs else (0.2, 99.8)
    angle = kwargs['angle'] if 'angle' in kwargs else 60
    transformation_dict = {
        'blur': normalize(ndimage.uniform_filter(X)),
        'invert': normalize(util.invert(X)),
        'rotate': rotate(X, angle=angle),
        'rescale_intensity': _rescale_intensity(X, q=q),
        'gamma_correction': exposure.adjust_gamma(X, gamma=0.4, gain=0.9),
        'log_correction': exposure.adjust_log(X),
        'sigmoid_correction': exposure.adjust_sigmoid(X),
        'horizontal_flip': X[:, ::-1],
        'vertical_flip': X[::-1, :],
        'rgb2gray': skimage.color.rgb2gray(X)
    }
    return transformation_dict[method_type]


# Helper functions for filtering based on conditional type
greater_than = lambda X, value: np.where(X > value)
less_than = lambda X, value: np.where(X < value)
equal_to = lambda X, value: np.where(X == value)
greater_than_or_equal = lambda X, value: np.where(X >= value)
less_than_equal = lambda X, value: np.where(X < value)
in_between = lambda X, min_value, max_value: np.where((X >= min_value) & (X <= max_value))


def flip_pixels(X, num_of_pixel, filtered_pixel=None, replace_with=0, random_state=0):
    # make a deep copy of the original image
    import copy
    X_modified = copy.deepcopy(X)
    np.random.seed(random_state)
    try:
        if len(filtered_pixel) > 0 & isinstance(filtered_pixel, tuple):
            f_pixels = filtered_pixel
            logger.info("Number of pixels matching the condition : {}".format(len(f_pixels[0])))
            logger.info("Number of pixels specified to be replaced : {}".format(num_of_pixel))

            if len(f_pixels) == 3:
                # uniformly random
                h = np.random.choice(f_pixels[0], num_of_pixel, replace=False)
                w = np.random.choice(f_pixels[1], num_of_pixel, replace=False)
                c = np.random.choice(f_pixels[2], num_of_pixel, replace=False)

                # for the selected pixels, set the pixel intensity to 0
                for h_i, w_i, c_i in zip(h, w, c):
                    X_modified[h_i, w_i, c_i] = replace_with
            elif len(f_pixels) == 2:
                # uniformly random
                h = np.random.choice(f_pixels[0], num_of_pixel, replace=False)
                w = np.random.choice(f_pixels[1], num_of_pixel, replace=False)

                # for the selected pixels, set the pixel intensity to 0
                for h_i, w_i in zip(h, w):
                    X_modified[h_i, w_i] = replace_with
            else:
                logger.info("Ambiguity in the shape of the input image : {}".format(X.shape))
    except:
        logger.info("No matching pixel for the specified condition")
    return X_modified


def flip_orientation(X, axis=0):
    if not isinstance(X, np.ndarray):
        err_msg = "Input array should be numpy.ndarray"
        raise(DataSetError(err_msg))

    return np.flip(X, axis)


def normalize(X):
    """ Normalize image of the shape (H, W, D) in the range of 0 and 1 (Min-Max scaling)
    """
    if not isinstance(X, np.ndarray):
        err_msg = "Input array should be numpy.ndarray"
        raise(DataSetError(err_msg))

    return np.array((X - np.min(X)) / (np.max(X) - np.min(X)))


def standard_scaler(X):
    if not isinstance(X, np.ndarray):
        err_msg = "Input array should be numpy.ndarray"
        raise(DataSetError(err_msg))

    X = (X - X.mean()) / X.std()
    return X


def show_image(X, cmap=None, bins=None, intensity_plot=False, width=8, height=8, title='Original'):
    # TODO: Add ability to handle a pre-defined axes for plotting
    import copy
    font = {'family': 'avenir',
            'color': 'black',
            'weight': 'normal',
            'size': 14,
            }
    _X = copy.deepcopy(X)

    try:
        import matplotlib.pyplot as plt
        n_clmns = 2 if intensity_plot else 1
        fig, ax = plt.subplots(1, n_clmns, figsize=(width, height))
        if intensity_plot:
            ax[0].imshow(_X, cmap=cmap)
            ax[0].set_title(title)
            ax[1].hist(_X.ravel(), bins=bins, histtype='step')
            ax[1].set_xlabel('Pixel intensity', fontdict=font)
            fig.subplots_adjust(wspace=0.3)
        else:
            ax.imshow(_X, cmap=cmap)
            ax.set_title(title)
    except ImportError:
        raise (MatplotlibUnavailableError("Matplotlib is required but unavailable on the system."))


padding_3d = lambda X, n_r, n_col, n_c: np.pad(X, ((0, n_r), (0, n_col), (0, n_c)), mode='constant', constant_values=0)
padding_2d = lambda X, n_r, n_col: np.pad(X, ((0, n_r), (0, n_col)), mode='constant', constant_values=0)


def view_windows(X, window_shape, steps):
    # Reference: https://codereview.stackexchange.com/questions/78156/moving-window-with-complete-boundary-in-python
    from skimage.util.shape import view_as_windows
    # number of rows to expand
    n_rows = np.mod(X.shape[0], steps)
    # number of columns to expand
    n_cols = np.mod(X.shape[1], steps)
    n_channel = np.mod(X.shape[2], steps) if len(window_shape) == 3 else None
    X_extended = padding_3d(X, n_rows, n_cols, n_channel) if n_channel is not None else padding_2d(X, n_rows, n_cols)
    rolling_view = view_as_windows(X_extended, window_shape, steps)
    return rolling_view



# -*- coding: UTF-8 -*-

from sklearn.ensemble import RandomForestClassifier
from sklearn.svm import LinearSVC
from sklearn.calibration import CalibratedClassifierCV
from sklearn.exceptions import NotFittedError
from sklearn.utils import shuffle
import pandas as pd
import numpy as np

from skater.core.global_interpretation.interpretable_models.brlc import BRLC
from skater.util import exceptions


class BigDataBRLC(BRLC):
    """
    :: Experimental :: The implementation is currently experimental and might change in future

    BigDataBRLC is a BRLC to handle large data-sets. Advisable to be used when the number of
    input examples>1k. It approximates large datasets with the help of surrogate(metamodel) estimators. For example, it uses
    surrogate estimator such as SVC(Support Vector Classifier) or RandomForest by default to filter the data
    points which are closest to the decision boundary. The idea is to identify the minimum training set size
    (controlled by the parameter sub_sample_percentage) with the goal to maximize accuracy.
    This helps in reducing the computation time to build the final BRL.

    Parameters
    ----------
    sub_sample_percentage : float (default=0.1)
        specify the fraction of the training sample to be retained for training BRL.

    iterations : int (default=30000)
        number of iterations for each MCMC chain.

    pos_sign : int (default=1)
        sign for the positive labels in the "label" column.

    neg_sign : int (default=0)
        sign for the negative labels in the "label" column.

    min_rule_len : int (default=1)
        minimum number of cardinality for rules to be mined from the data-frame.

    max_rule_len : int (default=8)
        maximum number of cardinality for rules to be mined from the data-frame.

    min_support_pos : float (default=0.1)
        a number between 0 and 1, for the minimum percentage support for the positive observations.

    min_support_neg : float (default 0.1)
        a number between 0 and 1, for the minimum percentage support for the negative observations.

    eta : int (default=1)

    n_chains: int (default=10)

    alpha : int (default=1)
        a prior pseudo-count for the positive(alpha1) and negative(alpha0) classes. Default values (1, 1)

    lambda_ : int (default=8)
        a hyper-parameter for the expected length of the rule list.

    discretize : bool (default=True)
        apply discretizer to handle continuous features.

    drop_features : bool (default=False)
        once continuous features are discretized, use this flag to either retain or drop them from the dataframe

    threshold : float (default=0.5)
        specify the threshold for the decision boundary. This is the probability level to compute \
        distance of the predictions(for input examples) from the decision boundary. Input examples closest to the \
        decision boundary are sub-sampled. Size of sub-sampled data is controlled using 'sub_sample_percentage'.

    penalty_param_svm : float (default=0.01)
        Regularization parameter('C') for Linear Support Vector Classifier. Lower regularization value forces the \
        optimizer to maximize the hyperplane.

        References: https://stats.stackexchange.com/questions/31066/what-is-the-influence-of-c-in-svms-with-linear-kernel

    calibration_type : string (default='sigmoid')
        Calibrate the base estimator's prediction(currently, all the base estimators are calibrated, that might
        change in future with more experimentation). Calibration could be performed in 2 ways
        1. parametric approach using Platt Scaling ('sigmoid')
        2. non-parametric approach using isotonic regression('isotonic).
        Avoid using isotonic regression for input examples<<1k because it tends to over-fit.

        References:

        [1] A. Niculescu-Mizil & R. Caruana(ICML2005), Predicting Good Probabilities With Supervised Learning

        [2] https://www.cs.cornell.edu/~alexn/papers/calibration.icml05.crc.rev3.pdf

        [3] http://fastml.com/classifier-calibration-with-platts-scaling-and-isotonic-regression/

    cv_calibration : int (default=3)
        specify number of folds for cross-validation splitting strategy

    random_state: int (default=0)
    surrogate_estimator: string (default='SVM', 'RF': RandomForest)
        Surrogate model to build the initial model for handling large datasets. Currently, SVM and RandomForest
        is supported.

    References
    ----------

    .. [1] Dr. Tamas Madl, https://github.com/tmadl/sklearn-expertsys/blob/master/BigDataRuleListClassifier.py
    .. [2] https://pdfs.semanticscholar.org/e44c/9dcf90d5a9a7e74a1d74c9900ff69142c67f.pdf
    .. [3] Surrogate model: https://en.wikipedia.org/wiki/Surrogate_model
    .. [4] W. Andrew Pruett , Robert L. Hester(2016),
       The Creation of Surrogate Models for Fast Estimation of Complex Model Outcomes
       (http://journals.plos.org/plosone/article?id=10.1371/journal.pone.0156574)

    Examples
    --------
    >>> from skater.core.global_interpretation.interpretable_models.brlc import BRLC
    >>> from skater.core.global_interpretation.interpretable_models.bigdatabrlc import BigDataBRLC
    >>> import pandas as pd
    >>> from sklearn.model_selection import train_test_split
    ...
    >>> Xtrain, Xtest, ytrain, ytest = train_test_split(X, y)
    >>> input_df = pd.read_csv('input_data.csv', skiprows=1)
    >>> sbrl_big = BigDataBRLC(sub_sample_percentage=0.1, min_rule_len=1, max_rule_len=3, iterations=10000,
    ...                                                 n_chains=3, surrogate_estimator="SVM", drop_features=True)
    >>> n_x, n_y = sbrl_big.subsample(Xtrain, ytrain, pos_label=1)
    >>> model = sbrl_big.fit(n_x, n_y, bin_labels='default')
    # For a complete example refer to credit_analysis_rule_lists.ipynb notebook in the `examples` section
    """
    def __init__(self, sub_sample_percentage=0.1, iterations=30000, pos_sign=1, neg_sign=0, min_rule_len=1,
                 max_rule_len=8, min_support_pos=0.10, min_support_neg=0.10, eta=1.0, n_chains=10, alpha=1,
                 lambda_=8, discretize=True, drop_features=False, threshold=0.5, penalty_param_svm=0.01,
                 calibration_type='sigmoid', cv_calibration=3, random_state=0, surrogate_estimator='SVM'):

        self.sample_percentage = sub_sample_percentage
        self.threhold = threshold

        self.surrogate_estimator = CalibratedClassifierCV(RandomForestClassifier(warm_start=True, oob_score=True,
                                                                                 max_features="sqrt",
                                                                                 random_state=random_state),
                                                          method=calibration_type, cv=cv_calibration) \
            if surrogate_estimator is 'RF' else surrogate_estimator

        self.surrogate_estimator = CalibratedClassifierCV(LinearSVC(C=penalty_param_svm, random_state=random_state),
                                                          method=calibration_type, cv=cv_calibration) \
            if surrogate_estimator is 'SVM' else surrogate_estimator

        super(BigDataBRLC, self).__init__(iterations, pos_sign, neg_sign, min_rule_len,
                                          max_rule_len, min_support_pos, min_support_neg, eta, n_chains, alpha,
                                          lambda_, discretize, drop_features)


    def subsample(self, X, y, pos_label=1, neg_label=0):
        """ subsampler to filter the input examples closer to the decision boundary

        Parameters
        -----------
        X : pandas.DataFrame
            input examples representing the training set
        y : pandas.DataFrame
            target labels associated with the training set
        pos_label : int
        neg_label : int

        Returns
        --------
        X_, y_ : pandas.dataframe
        sub-sampled input examples
        """
        if not isinstance(X, pd.DataFrame) or not isinstance(y, pd.Series):
            raise exceptions.DataSetError("Only pandas.DataFrame as input type is currently supported")

        # validate the consistency of the input data
        if not X.shape[0] == y.shape[0]:
            raise exceptions.DataSetError("mismatch in the shape of X and y")

        try:
            self.surrogate_estimator.predict_proba(X[0:1])
        except NotFittedError:
            self.surrogate_estimator.fit(X, y)

        est_prob_scores = pd.DataFrame(self.surrogate_estimator.predict_proba(X))

        # compute the distance from the decision boundary
        distance_from_threshold = est_prob_scores[pos_label].apply(lambda x: np.abs(self.threhold - x))
        pos_label_index = np.where(y == pos_label)[0]
        neg_label_index = np.where(y == neg_label)[0]

        pos_label_dist = distance_from_threshold[pos_label_index]
        neg_label_dist = distance_from_threshold[neg_label_index]

        # sort the neighboring distances from the threshold in the ascending order to select points which
        # are closer to the decision boundary
        sorted_dist_pos_label = pos_label_dist.sort_values()
        sorted_dist_neg_label = neg_label_dist.sort_values()

        # sub-sample the data
        number_of_rows = len(y) * self.sample_percentage
        pos_fraction = len(pos_label_index) / float(len(y))
        neg_fraction = 1 - pos_fraction

        pos_df = pd.DataFrame(X.iloc[sorted_dist_pos_label[:int(number_of_rows * pos_fraction) + 1].index])
        neg_df = pd.DataFrame(X.iloc[sorted_dist_neg_label[:int(number_of_rows * neg_fraction) + 1].index])

        new_X = pd.concat([pos_df, neg_df], axis=0)

        # Randomly shuffle the newly formed data-set
        X_, y_ = shuffle(new_X, y[new_X.index])
        return X_, y_



# -*- coding: UTF-8 -*-
import warnings
import numpy as np
from tensorflow.python.ops import nn_grad, math_grad

from skater.util.logger import build_logger
from skater.util.logger import _INFO


class Initializer(object):
    """
    """
    __name__ = "Initializer"
    # Currently supported Activation ops
    activation_ops = ['Relu', 'Elu', 'Softplus', 'Tanh', 'Sigmoid']
    _enabled_method_class = None
    _grad_override_checkflag = 0

    logger = build_logger(_INFO, __name__)

    def __init__(self, output_tensor, input_tensor, samples, session):
        self.output_tensor = output_tensor
        self.input_tensor = input_tensor
        self.samples = samples
        self.session = session


    def _session_run(self, output_tensor, samples):
        feed_dict = {}
        feed_dict[self.input_tensor] = samples
        return self.session.run(output_tensor, feed_dict)


    def _validate_baseline(self, baseline):
        if baseline is not None and baseline.shape != ((1,) + self.samples.shape[1:]):
            if baseline.shape == self.samples.shape[1:]:
                baseline = np.expand_dims(baseline, 0)
            else:
                raise RuntimeError('Baseline input shape {} does not match expected input shape {}'
                                   .format(baseline.shape, self.samples.shape[1:]))
        elif baseline is None:
            baseline = np.zeros((1,) + self.samples.shape[1:])
        return baseline


    @classmethod
    def _original_grad(cls, op, grad):
        if op.type not in cls.activation_ops:
            warnings.warn('Selected Activation Ops({}) is currently not supported.'.format(op.type))
        op_name = '_{}Grad'.format(op.type)
        Initializer.logger.debug("Operation name : {}".format(op_name))
        ops_func = getattr(nn_grad, op_name) if hasattr(nn_grad, op_name) else getattr(math_grad, op_name)
        return ops_func(op, grad)



"""Includes kernels, static type/properties for package, exceptions, logging."""


__all__ = [
    'rbf_kernel',
    'StaticTypes',
    'return_data_type',
    'build_static_predictor',
    'ModelTypes',
    'OutputTypes',
    'ControlledDict',
    'pickle_method',
    'unpickle_method',
    'plot_from_df']



from sklearn.datasets import load_breast_cancer
from sklearn.neighbors import KNeighborsClassifier
from skater.model import InMemoryModel
from skater.util.static_types import StaticTypes
from skater.tests.arg_parser import arg_parse, create_parser
import unittest


class TestModel(unittest.TestCase):
    """
    Tests the skater.model.* types
    """

    def setUp(self):
        """Create data for testing"""
        self.parser = create_parser()
        args = self.parser.parse_args()
        debug = args.debug

        if debug:
            self.log_level = 10
        else:
            self.log_level = 30


    def test_issues_161_and_189(self):
        """
        ensure DataManager(data).data == data
        """
        X, y = load_breast_cancer(True)
        X, y = X[15:40], y[15:40]
        model = KNeighborsClassifier(weights='distance', p=2, n_neighbors=10).fit(X, y)
        skater_model = InMemoryModel(model.predict_proba, examples=X, probability=True)
        assert skater_model.probability is True
        assert skater_model.model_type == StaticTypes.model_types.classifier


if __name__ == '__main__':
    runner = unittest.TextTestRunner(verbosity=2)
    runner.run(unittest.makeSuite(TestModel))



import unittest

import pandas as pd
from sklearn.datasets import make_moons
from sklearn.tree import DecisionTreeClassifier
from sklearn.model_selection._split import train_test_split

from skater.core.explanations import Interpretation
from skater.model import InMemoryModel
from skater.util.logger import _INFO


class TestTreeSurrogates(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        # Classification use-case
        cls.X_c, cls.y_c = make_moons(1000, noise=0.5)
        cls.X_c = pd.DataFrame(cls.X_c, columns=['F1', 'F2'])
        cls.target_names = ['class 0', 'class 1']
        cls.X_train_c, cls.X_test_c, cls.y_train_c, cls.y_test_c = train_test_split(cls.X_c, cls.y_c)
        cls.classifier_est = DecisionTreeClassifier(max_depth=5, random_state=5)
        cls.classifier_est.fit(cls.X_train_c, cls.y_train_c)
        cls.interpreter = Interpretation(cls.X_train_c, feature_names=cls.X_c.columns)
        cls.model_inst = InMemoryModel(cls.classifier_est.predict, examples=cls.X_train_c,
                                       model_type='classifier', unique_values=[0, 1], feature_names=cls.X_c.columns,
                                       target_names=cls.target_names, log_level=_INFO)

    # all the below tests are with F1-score
    def test_surrogate_no_pruning(self):
        surrogate_explainer = self.interpreter.tree_surrogate(oracle=self.model_inst, seed=5)
        result = surrogate_explainer.fit(self.X_train_c, self.y_train_c, use_oracle=True,
                                         prune=None, scorer_type='default')
        self.assertEquals(result < 0, True)


    def test_surrogate_with_prepruning(self):
        surrogate_explainer = self.interpreter.tree_surrogate(oracle=self.model_inst, seed=5)
        result = surrogate_explainer.fit(self.X_train_c, self.y_train_c, use_oracle=True,
                                         prune='pre', scorer_type='f1')
        self.assertEquals(result < 0, True)


    def test_surrogate_with_postpruning(self):
        surrogate_explainer = self.interpreter.tree_surrogate(oracle=self.model_inst, seed=5)
        result = surrogate_explainer.fit(self.X_train_c, self.y_train_c, use_oracle=True,
                                         prune='post', scorer_type='f1')
        self.assertEquals(result < 0, True)


    def test_surrogate_with_cross_entropy(self):
        model_inst = InMemoryModel(self.classifier_est.predict_proba, examples=self.X_train_c,
                                   model_type='classifier', feature_names=self.X_c.columns,
                                   target_names=self.target_names, log_level=_INFO, probability=True)
        surrogate_explainer = self.interpreter.tree_surrogate(oracle=model_inst, seed=5)
        result = surrogate_explainer.fit(self.X_train_c, self.y_train_c, use_oracle=True,
                                         prune='post', scorer_type='default')
        self.assertEqual(surrogate_explainer.scorer_name_, 'cross-entropy', True)
        self.assertEquals(result != 0, True)


if __name__ == '__main__':
    suite = unittest.TestLoader().loadTestsFromTestCase(TestTreeSurrogates)
    unittest.TextTestRunner(verbosity=2).run(suite)



from matplotlib.cm import get_cmap
import matplotlib as mpl
from matplotlib.patches import Patch
import pandas as pd

from skater.data.datamanager import DataManager as DM
from skater.util.exceptions import MatplotlibUnavailableError
from skater.util.logger import build_logger
from skater.util.logger import _INFO
from skater.core.local_interpretation.text_interpreter import relevance_wt_assigner
from skater.util.dataops import convert_dataframe_to_dict
from skater.util.text_ops import generate_word_list

logger = build_logger(_INFO, __name__)


def __set_plot_feature_relevance_keyword(**plot_kw):
    plot_name = plot_kw['plot_name'] if 'plot_name' in plot_kw.keys() else 'feature_relevance.png'
    top_k = plot_kw['top_k'] if 'top_k' in plot_kw.keys() else 10
    color_map = plot_kw['color_map'] if 'color_map' in plot_kw.keys() else ('Red', 'Blue')
    fig_size = plot_kw['fig_size'] if 'fig_size' in plot_kw.keys() else (20, 10)
    font_name = plot_kw['font_name'] if 'font_name' in plot_kw.keys() else "Avenir Black"
    txt_font_size = plot_kw['txt_font_size'] if 'txt_font_size' in plot_kw.keys() else '14'
    return plot_name, top_k, color_map, fig_size, font_name, txt_font_size


# Reference: https://stackoverflow.com/questions/30618002/static-variable-in-a-function-with-python-decorator
def static_var(varname, value):
    def decorate(func):
        setattr(func, varname, value)
        return func
    return decorate


# The counter is added for a weird caching error that occurs when including images to html
# https://stackoverflow.com/questions/1431512/is-there-a-way-to-force-browsers-to-refresh-download-images
@static_var("plot_counter", 0)
def build_visual_explainer(text, relevance_scores, font_size='12pt', file_name='rendered', title='Word Relevance',
                           pos_clr_name='Reds', neg_clr_name='Blues', alpha=0.7, highlight_oov=False,
                           enable_plot=False, **plot_kw):
    """
    Build a visual explainer highlighting the positive(color code: Red(default)) and
    negative(color code: Blue(default)) relevance of the input features

    Parameters
    ----------
    text: str
        input text to be explained
    relevance_scores: pandas.core.series.Series
        contribution/attribution scores for the features corresponding to the given text.
        The relevance scores are compute over the embedding vector, remember to aggregate the embedding dimension
        by computing 'mean/sum' to get scalar coefficient for the features.
    font_size: str (default: `12pt`)
        font size for the text visualizer
    file_name: str (default: 'rendered')
        specify the file name for the html file to be generated
    title: str
        specify a title
    pos_clr_name: str (default: 'Reds')
        color name for highlight positive influence. Color names should be compatible with matplotlib supported color names
        (https://matplotlib.org/gallery/color/named_colors.html)
    neg_clr_name: str (default: 'Blues')
        color name for highlight negative influence. Color names should be compatible with matplotlib supported color names
        (https://matplotlib.org/gallery/color/named_colors.html)
    alpha: int (default: 0.7)
        adjust the transparency for display
    highlight_oov: boolean (default: False)
        highlighting words which are not part of the dictionary
    enable_plot: boolean (default: False)
        if True, generate feature relevance plots as well
    plot_kw: optional
        optional aesthetic features for plotting,
        - plot_name: str (default: feature_relevance.png)
        - top_k: int (default: 10)
        - color_map: tuple of str('positive_color', 'negative_color') (default: ('Red', 'Blue'))
        - fig_size: tuple of int(width, height) (default: (20, 10)
        - font_name: str (default: 'Avenir Black')
        - txt_font_size: str( default: '14')

    Returns
    -------
     None

    """
    # TODO: Add support for better visualization and plotting frameworks- e.g bokeh, plotly
    # Process the raw text to a word list
    _words = generate_word_list(text, ' ')
    features_df = pd.DataFrame({'features': _words})
    scores_df = pd.DataFrame({'relevance_scores': relevance_scores.tolist()})
    # Merge the data-frame column-wise, assert if the length of the word list and relevance
    # score data-frame does not match. This is important as currently the join is done on the index
    assert len(_words) == scores_df.shape[0]
    words_scores_df = features_df.join(scores_df)
    # assign column names to df containing 'words | relevance score'
    words_scores_df.columns = ['features', 'relevance_scores']
    words_scores_dict = convert_dataframe_to_dict('features', 'relevance_scores', words_scores_df)

    # generate plot displaying feature relevance rank ordered
    f_name = plot_feature_relevance(words_scores_df, **plot_kw) if enable_plot else None
    logger.info("Relevance plot name: {}".format(f_name))

    # build the html structure
    h_str = _build_str(text, words_scores_dict, f_name, title, font_size, pos_clr_name,
                       neg_clr_name, alpha, highlight_oov)
    # build html
    _build_html_file(html_str=h_str, file_name=file_name)


def _build_str(text, words_scores_dict, plot_file_name, title, font_size,
               pos_clr_name, neg_clr_name, alpha_value, highlight_oov):
    """
    References
    ----------
    ..  [1] https://github.com/cod3licious/textcatvis/blob/master/textcatvis/vis_utils.py
    ..  [2] http://matplotlib.org/examples/color/colormaps_reference.html
    """
    # TODO: Add a HTML validator to verify that HTML structure
    # color-maps for the words
    cmap_pos = get_cmap(pos_clr_name)
    cmap_neg = get_cmap(neg_clr_name)
    # Highlight with 'yellow' if the word is not present in
    # the word dictionary or not contributing to actual prediction
    rgba = (1., 1., 0)
    norm = mpl.colors.Normalize(0., 1.)

    html_content = u'<body><h3>{}</h3>' \
                   u'<div class="row" style=background-color:#F5F5F5 ' \
                   u'white-space: pre-wrap; ' \
                   u'font-size: {}; ' \
                   u'font-family: Avenir Black>'.format(title, font_size)

    rest_text = text
    relevance_wts = relevance_wt_assigner(text, words_scores_dict)
    for word, wts in relevance_wts:
        html_content += rest_text[: rest_text.find(word)]
        # cut off the identified word
        rest_text = rest_text[rest_text.find(word) + len(word):]
        # adjust opacity for non-vocabulary words
        alpha = 0.5 if highlight_oov else 0.
        if wts is not None:
            # override the RGB values for color coding
            rgba = cmap_neg(norm(-wts)) if wts < 0 else cmap_pos(norm(wts))
            # adjusting opacity for in-dictionary words
            alpha = alpha_value
        html_content += u'<span style="background-color: rgba({:.0f}, {:.0f}, {:.0f}, {:.1f})">{}</span>'\
            .format(round(float(255) * rgba[0]), round(float(255) * rgba[1]), round(float(255) * rgba[2]), alpha, word)
    # rest of the text
    html_content += rest_text
    html_content += u'</div>'

    # Embed the feature relevance scores as a plot
    build_visual_explainer.plot_counter += 1
    html_content += u'<div align="center"><img src="./{}?{}"</div>'. \
        format(plot_file_name, build_visual_explainer.plot_counter) if plot_file_name is not None else ''
    html_content += u'</body>'
    return html_content


def _build_html_file(html_str, file_name):
    file_name_with_extension = '{}.html'.format(file_name)
    with open(file_name_with_extension, 'w', encoding='utf8') as f:
        f.write(html_str)
        logger.info("Visual Explainer built, "
                    "use show_in_notebook to render in Jupyter style Notebooks: {}".format(file_name_with_extension))


def plot_feature_relevance(feature_relevance_scores, **plot_kw):
    """ Plotting function for visualizing the feature contribution rank ordered wrt +ve/-ve effect

    Parameters
    ----------
    feature_relevance_scores: pandas.core.frame.DataFrame
    plot_kw: optional
        optional aesthetic features for plotting,
        - plot_name: str (default: feature_relevance.png)
        - top_k: int (default: 10)
        - color_map: tuple of str('positive_color', 'negative_color') (default: ('Red', 'Blue'))
        - fig_size: tuple of int(width, height) (default: (20, 10)
        - font_name: str (default: 'Avenir Black')
        - txt_font_size: str( default: '14')

    Returns
    -------
    f_name: str
        returns the generate plot saved as *.png file. Other formats could be specified as well.
    """
    try:
        import matplotlib.pyplot as plt
    except ImportError:
        raise (MatplotlibUnavailableError("Matplotlib is required but unavailable on the system."))

    # Validate the input data format. Currently supported types include (pandas.DataFrame)
    DM._check_input(feature_relevance_scores)
    # set the params needed for the plot
    f_name, top_k, color_map, fig_size, font_name, txt_font_size = __set_plot_feature_relevance_keyword(**plot_kw)

    # set the colors
    pos_color = color_map[0]
    neg_color = color_map[1]

    df = feature_relevance_scores.sort_values(by='relevance_scores', ascending=False)
    df['positive'] = df['relevance_scores'] > 0

    # Setting the style globally for the plot
    # For other style check the reference here `plt.style.available`
    plt.clf()
    plt.style.use('bmh')
    fig = plt.figure()
    ax = fig.add_subplot(111)
    # filter the top k features wrt each signed category (Positive / Negative ) features
    df_filtered = df.groupby('positive').head(top_k)
    color_list = df_filtered['positive'].map({True: pos_color, False: neg_color})
    df_filtered['relevance_scores'].plot(ax=ax, kind='barh', color=color_list, figsize=fig_size, width=0.85)
    custom_lines = [Patch(facecolor=pos_color, edgecolor='r'), Patch(facecolor=neg_color, edgecolor='b')]

    ax.legend(custom_lines, ['positive', 'negative'], loc='best')
    ax.yaxis.set_visible(False)
    # Display the feature names on the plot
    for index, f_x in enumerate(df_filtered['features']):
        ax.text(0, index + 0.5, f_x, ha='right', fontsize=txt_font_size, fontname=font_name)
    plt.savefig(f_name)
    logger.info("Rank order feature relevance based on input created and saved as {}".format(f_name))
    return f_name



import pandas as pd
import numpy as np
import unittest
import sys
try:
    if sys.version_info >= (3, 5):
        from skater.core.global_interpretation.interpretable_models.brlc import BRLC
        from skater.core.validation import compute_validation_curve
except ImportError as err:
    # Because of unresolved error related to ImportError: dlopen: cannot load any more object with static TLS(#229)
    raise unittest.SkipTest(str(err))


# When in Dev mode, a consistent mode to validate test, that would keep track of weird segmentation fault is using gdb
# (GNU Debugger). This is a temporary workaround. Follow the below mentioned steps
# 1. sudo apt install gdb
# 2. gdb python
# 3. r skater/tests/test_rule_list.py
# 4. Result: All the tests should succeed, and if it fails it will point out the trace for failure
@unittest.skipIf(sys.version_info < (3, 5), "only Bayesian Rule List supported only for python 3.5/3.6")
class TestRuleList(unittest.TestCase):

    @classmethod
    def setUpClass(cls):
        cls.sbrl_inst = BRLC(min_rule_len=1, max_rule_len=2, iterations=10000, n_chains=3)
        cls.input_data = pd.read_csv('skater/tests/data/sample_data.csv')
        # data transformation and cleaning ...
        cls.input_data["Sex"] = cls.input_data["Sex"].astype('category')
        cls.input_data["Sex_Encoded"] = cls.input_data["Sex"].cat.codes
        cls.input_data["Embarked"] = cls.input_data["Embarked"].astype('category')
        cls.input_data["Embarked_Encoded"] = cls.input_data["Embarked"].cat.codes
        cls.input_data = cls.input_data.drop(['Ticket', 'Cabin', 'Name', 'Sex', 'Embarked'], axis=1)
        # Remove NaN values
        cls.input_data = cls.input_data.dropna()
        cls.y = cls.input_data['Survived']
        cls.input_data = cls.input_data.drop(['Survived'], axis=1)
        # Train a model
        cls.sbrl_inst.fit(cls.input_data[1:50], cls.y[1:50], undiscretize_feature_list=["PassengerId", "Pclass",
                                                                                        "SibSp", "Parch",
                                                                                        "Sex_Encoded",
                                                                                        "Embarked_Encoded"])


    def test_discretizer(self):
        new_df = self.sbrl_inst.discretizer(self.input_data, column_list=["Age"])
        self.assertEquals(new_df["Age_q_label"].shape[0] > 0, True)


    def test_model_build(self):
        new_data = self.sbrl_inst.discretizer(self.input_data, column_list=["Age", "Fare"])
        result_score = self.sbrl_inst.predict_proba(new_data)
        result_labels = self.sbrl_inst.predict(new_data)

        # make sure shape of the data-frame is as expected
        self.assertEquals(result_score.shape, (77, 2))
        self.assertEquals(result_labels[1].shape, (77, ))

        generated_labels = np.unique(result_labels[1])
        expected_labels = np.array([0, 1])
        self.assertEquals(np.array_equal(generated_labels, expected_labels), True)


    def test_model_save_load(self):
        self.sbrl_inst.save_model("test.pkl", compress=True)
        # Explicitly assigning the model instance to 'None' to validate loading of persisted model
        # Care is advised when handing the model instance, it might make the model unstable
        self.sbrl_inst.model = None
        self.assertEquals(self.sbrl_inst.model is None, True)
        self.sbrl_inst.load_model("test.pkl")
        self.assertEquals(self.sbrl_inst.model is not None, True)


    def test_model_output(self):
        result = self.sbrl_inst.access_learned_rules('23:25')
        self.assertEquals(len(result), 2)


    @unittest.skip("Support for computing validation curve for SBRL is still under development")
    def test_validation(self):
        param_range = [3, 4]
        train_scores, test_scores = compute_validation_curve(self.sbrl_inst, n_folds=2, x=self.input_data, y=self.y,
                                                             param_name="rule_minlen", param_range=param_range)
        self.assertEquals(train_scores.shape[0], 2)
        self.assertEquals(test_scores.shape[0], 2)


if __name__ == '__main__':
        suite = unittest.TestLoader().loadTestsFromTestCase(TestRuleList)
        unittest.TextTestRunner(verbosity=2).run(suite)



import unittest
import pandas as pd
import numpy as np

from skater.util.text_ops import cleaner
from skater.util.dataops import convert_dataframe_to_dict
from skater.util.text_ops import generate_word_list


class TestTextOps(unittest.TestCase):

    def setUp(self):
        pass


    def test_cleaner(self):
        raw_text = "    11111this is just an example..., don't be surprised!!!0000000    "
        expected_result = "this is just an example , don t be surprised"
        result = cleaner(raw_text, to_lower=True, norm_num=False, char_to_strip="0|' '|1")
        self.assertEquals(result, expected_result)


    def test_dataframe_to_dict_converter(self):
        # converting of dataframe to dict
        data = np.array([['', 'feature_name', 'relevance_scores'],
                         ['R1', 'F1', 20],
                         ['R2', 'F2', 40]])
        test_df = pd.DataFrame(data=data[1:, 1:], index=data[1:, 0], columns=data[0, 1:])
        test_dict = convert_dataframe_to_dict("feature_name", "relevance_scores", test_df)
        self.assertEquals(test_dict['F1'], '20')


    def test_generate_word_list(self):
        input_str = "Summer is    coming and I like it"
        result = generate_word_list(input_str, " ")
        self.assertEquals(len(result), 7)


if __name__ == '__main__':
    suite = unittest.TestLoader().loadTestsFromTestCase(TestTextOps)
    unittest.TextTestRunner(verbosity=2).run(suite)



# Reference: http://python-packaging.readthedocs.io/en/latest/dependencies.html
from setuptools import setup, find_packages
from setuptools.command.install import install
import subprocess
import os
import io
import sys
import contextlib


class InvokingShScript(install):
    """Custom install setup to help run shell commands (outside shell) before installation"""

    # reference:
    # Syntax (long option, short option, description).
    user_options = install.user_options + [
        ('ostype=', None, 'The OS type of the box(linux-ubuntu/mac)'),
        ('rl=', None, 'Enable rule based learning')
    ]

    def initialize_options(self):
        install.initialize_options(self)
        self.ostype = 'linux-ubuntu'
        self.rl = False

    def finalize_options(self):
        if self.ostype is None:
            raise Exception("specify os type ...")
        if self.rl is None:
            raise Exception(" should based learning be enabled? ...")
        install.finalize_options(self)

    def run(self):
        # install rule based learners when asked
        if self.rl:
            dir_path = os.path.dirname(os.path.realpath(__file__))
            shell_script_path = os.path.join(dir_path, 'setup.sh')

            subprocess.check_output(['bash', shell_script_path, self.ostype])
        install.do_egg_install(self)


@contextlib.contextmanager
def chdir(new_dir):
    old_dir = os.getcwd()
    try:
        os.chdir(new_dir)
        sys.path.insert(0, new_dir)
        yield
    finally:
        del sys.path[0]
        os.chdir(old_dir)


def setup_package():
    root = os.path.abspath(os.path.dirname(__file__))

    with chdir(root):
        with io.open(os.path.join(root, 'skater', 'about.py'), encoding='utf8') as f:
            about = {}
            exec(f.read(), about)

        with io.open(os.path.join(root, 'description.rst'), encoding='utf8') as f:
            readme = f.read()

    setup(
        name=about['__title__'],
        packages=find_packages(),
        description=about['__summary__'],
        long_description=readme,
        author=about['__author__'],
        author_email=about['__email__'],
        version=about['__version__'],
        url=about['__uri__'],
        license=about['__license__'],
        cmdclass={'install': InvokingShScript},
        include_package_data=True,
        install_requires=[
            'scikit-learn>=0.18',
            'scikit-image==0.14',
            'pandas>=0.22.0',
            'ds-lime>=0.1.1.21',
            'requests',
            'multiprocess',
            'dill>=0.2.6',
            'wordcloud==1.3.1',
            'joblib==0.11',
            'Jinja2==2.10',
            'pydotplus==2.0.2',
            'bs4'],
        extras_require={
            'all': 'matplotlib'},
        zip_safe=False)


if __name__ == '__main__':
    setup_package()



import numpy as np
from sklearn.ensemble import RandomForestRegressor, RandomForestClassifier


def build_X(N=1000, dim=3):
    return np.random.normal(0, 10, size=(N, dim))


def build_y_regression(X, b):
    return np.dot(X, b)


def build_y_classifier(X, b, n_classes=2):
    y = build_y_regression(X, b)
    N = len(X)
    n_per_class = int(N / n_classes)
    sorted_inds = y.argsort()

    result = np.zeros(N)

    for i in range(n_classes):
        inds = sorted_inds[i * n_per_class: (i + 1) * n_per_class]
        result[inds] = i

    return np.array(result)


def build_X_y_model(N=1000, dim=3, model_type='classifier', n_classes=2, beta=None):
    if beta is None:
        beta = np.random.normal(0, 1, dim)
    else:
        dim = len(beta)
    X = build_X(N, dim)
    if model_type == 'classifier':
        y = build_y_classifier(X, beta, n_classes=n_classes)
        model = RandomForestClassifier()
        model.fit(X, y)
        return X, y, model
    if model_type == 'regressor':
        y = build_y_regression(X, beta)
        model = RandomForestRegressor()
        model.fit(X, y)
        return X, y, model
